'''function to estimate the initial concentration of water on particles and walls'''

import numpy as np
from partit_var import kimt_calc
import sys
import matplotlib.pyplot as plt
import scipy.constants as si

def init_water_partit(x, y, H2Oi, Psat, mfp, siz_str, num_sb, num_speci, 
		accom_coeff, y_mw, surfT, R_gas, TEMP, NA, y_dens, 
		N_perbin, RH, core_diss, Varr, Vbou, rbou, Vol0, MV,
		therm_sp, Cw, kgwt, act_coeff, wall_on, 
		partit_cutoff, Press, coll_dia, seedi):

	# inputs: ------------------------------------------------------
	# x - radius of particles per size bin (um)
	# Psat - saturation vapour pressure of components (molecules/cc (air))
	# siz_str - the size structure
	# num_sb - number of size bins
	# N_perbin - initial particle concentration (#/cc (air))
	# Vbou - volume bounds (um3)
	# rbou - radius bounds (um)
	# MV - molar volume of components (cc/mol)
	# therm_sp - thermal speed of components (m/s) (num_speci)
	# Cw - concentration of wall (molecules/cc (air))
	# kgwt - mass transfer coefficient for vapour-wall partitioning (/s)
	# act_coeff - activity coefficients of components (dimensionless)
	# wall_on - marker for whether to consider wall
	# partit_cutoff - product of vapour pressure and activity coefficient
	#		at which gas-particle partitioning assumed zero (Pa)
	# Press - pressure inside chamber (Pa)
	# coll_dia - collision diameters of components (cm)
	# seedi - index of seed components
	# --------------------------------------------------------------
	

	if sum(N_perbin)>0.0: # if seed particles present
		# new array of size bin radii (um)
		print('Equilibrating water in vapour with water in seed particles')

		for sbstep in range(len(x)): # loop through size bins
			if N_perbin[sbstep]<1.0e-10: # no need to partition if no particle present
				continue
			
			# partitioning coefficient and kelvin factor
			[kimt, kelv_fac] = kimt_calc(y, mfp, num_sb, num_speci, accom_coeff, y_mw,   
				surfT, R_gas, TEMP, NA, y_dens, N_perbin, 
				x.reshape(1, -1)*1.0e-6, Psat, therm_sp, 
				H2Oi, act_coeff, wall_on, 0, partit_cutoff, Press, coll_dia)

			# get seed particle properties: concentration (molecules/cc (air))
			ycore = 0.			
			for ci in range(len(seedi)):
				ycore += y[num_speci*(sbstep+1)+seedi[ci]]*core_diss[ci]
			
			# first guess of particle-phase water concentration based on RH = mole 
			# fraction (molecules/cc (air))
			y[num_speci*(sbstep+1)+H2Oi] = (ycore/kelv_fac[sbstep])*RH
			
			# gas phase concentration of water (molecules/cc (air)), will stay constant
			# because RH is constant
			Cgit = y[H2Oi]
			
			# concentration of water at particle surface in gas phase (molecules/cc (air))
			Wc_surf = y[num_speci*(sbstep+1)+H2Oi]
			
			# total particle surface gas-phase concentration of all components 
			# (molecules/cc (air)):
			conc_sum = (np.sum(y[num_speci*(sbstep+1):num_speci*(sbstep+2)]))
			# account for seed dissociation constant
			for ci in range(len(seedi)):
				conc_sum = (conc_sum-y[num_speci*(sbstep+1)+seedi[ci]]+
					(y[num_speci*(sbstep+1)+seedi[ci]]*core_diss[ci]))
			
			# partitioning coefficient and kelvin factor
			[kimt, kelv_fac] = kimt_calc(y, mfp, num_sb, num_speci, accom_coeff, y_mw,   
					surfT, R_gas, TEMP, NA, y_dens, N_perbin, 
					x.reshape(1, -1)*1.0e-6, Psat, therm_sp, 
					H2Oi, act_coeff, wall_on, 0, partit_cutoff, Press, coll_dia)

			Csit = (Wc_surf/conc_sum)*Psat[0, H2Oi]*kelv_fac[sbstep, 0]
			
			y0 = np.zeros(len(y))
			y0[:] = y[:] # initial concentration before moving centre (molecules/cc (air))
			y00 = np.zeros(len(y))
			y00[:] = y[:] # initial concentration before iteration (molecules/cc (air))
			dydtfac = 1.0e-1
			dydt0 = 0 # change in water concentration (molecules/cc) on previous iteration
			
			while np.abs(Cgit-Csit)>Cgit/1.0e6:
				
				y0[:] = y[:] # update initial concentrations (molecules/cc (air))
				
				# new particle size, required to update kelvin factor and partitioning
				# coefficient in kimt_calc below 
				# number of moles of each component in a single particle (mol/cc (air))
				nmolC= ((y[num_speci*(sbstep+1):num_speci*(sbstep+2)]/(si.Avogadro*N_perbin[sbstep])))
				# new radius of single particle per size bin (um) including volume of 
				# water
				Vnew = np.sum(nmolC*MV*1.e12) # new volume (um3)
				x[sbstep] = ((3.0/(4.0*np.pi))*Vnew)**(1.0/3.0)
				
				# update partitioning coefficients
				[kimt, kelv_fac] = kimt_calc(y, mfp, num_sb, num_speci, accom_coeff, y_mw,   
					surfT, R_gas, TEMP, NA, y_dens, N_perbin, 
					x.reshape(1, -1)*1.0e-6, Psat, therm_sp, 
					H2Oi, act_coeff, wall_on, 0, partit_cutoff, Press, coll_dia)
				
				# concentration of water at particle surface in gas phase 
				# (molecules/cc (air))
				Wc_surf = y[num_speci*(sbstep+1)+H2Oi]
				
				# total particle surface gas-phase concentration of all species 
				# (molecules/cc (air)):
				conc_sum = (np.sum(y[num_speci*(sbstep+1):num_speci*(sbstep+2)]))
				
				# account for seed dissociation constant
				for ci in range(len(seedi)):
					conc_sum = (conc_sum-y[num_speci*(sbstep+1)+seedi[ci]]+
						(y[num_speci*(sbstep+1)+seedi[ci]]*core_diss[ci]))
					
				Csit = (Wc_surf/conc_sum)*Psat[0, H2Oi]*kelv_fac[sbstep, 0]*act_coeff[0, H2Oi]
		
				
				dydt = kimt[sbstep, H2Oi]*(Cgit-Csit) # partitioning rate (molecules/cc.s)
				dydtn = dydt*(dydtfac*RH*x[sbstep]**3.0)
				dydtn = dydt*(dydtfac*RH*x[sbstep]**3.0)/kelv_fac[sbstep, 0]
				
				# new estimate of concentration of condensed water (molecules/cc (air))
				y[num_speci*(sbstep+1)+H2Oi] += dydtn
				# if iteration becomes unstable, reset and reduce change per step
				if ((np.sum(y[num_speci*(sbstep+1):num_speci*(sbstep+2)]))<0.0):
					# return to first guess before iteration on this size bin began
					y[:] = y00[:]
					dydtfac = dydtfac/2.0
				# if iteration oscillates, reduce change per step
				if dydt0<0 and dydtn>0:
					dydtfac = dydtfac/2.0
				
				# remember change in this step
				dydt0 = dydtn
				
			
			sbstep += 1

		# call on the moving centre method for redistributing particles that have grown 
		# beyond their upper size bin boundary due to water condensation, note, only do this
		# after the iteration per size bin when we know the new particle-phase concentration 
		# of water
		# concentration in particles now (molecules/cc (air))
		Cp = y[num_speci:(num_speci*(num_sb-wall_on+1))]
		Cp = np.transpose(Cp.reshape((num_sb-wall_on), num_speci))
	
		if (siz_str == 0): # moving-centre size structure
			import mov_cen_water_eq
			
			(N_perbin, Varr, Cp, x, redt, blank, 
				tnew) = mov_cen_water_eq.mov_cen_main(N_perbin, Vbou, 
				Cp, (num_sb-wall_on), num_speci, Vol0, 0.0,
				0, MV*1.e12)
			
			if (redt == 1): # check on whether exception raised by moving centre
				print('Error whilst equilibrating seed particles with water vapour (inside init_water_partit module).  Please investigate, perhaps by checking rh and pconc inputs in model variables input file.  See README for guidance and how to report bugs.')
				sys.exit()

		if (siz_str == 1): # full-moving size structure
			import fullmov
			(Varr, x, Cp, N_perbin, Vbou, rbou) = fullmov.fullmov((num_sb-wall_on), N_perbin, num_speci, y[num_speci:(num_speci*(num_sb-wall_on+1))], MV*1.e12, Vol0, Vbou, rbou)

		# new particle-phase concentrations (molecules/cc (air))
		y[num_speci:(num_speci*(num_sb-wall_on+1))] = Cp

	if kgwt>1.0e-10 and Cw>0.0:
		print('Equilibrating water in vapour with water on wall')
	
		# first guess of wall-phase water concentration based on RH = mole 
		# fraction (molecules/cc (air))
		y[num_speci*num_sb+H2Oi] = Cw*RH
		

		# allow gas-phase water to equilibrate with walls
		while np.abs(y[H2Oi]-Psat[0, H2Oi]*((y[num_speci*(num_sb)+H2Oi]/Cw))*act_coeff[0, H2Oi])>1.0e2:
			
			# concentration of water at wall surface in gas phase (molecules/cc (air))
			Csit = y[num_speci*num_sb:num_speci*(num_sb+1)]
			Csit = (Psat[0, :]*(Csit/Cw)*act_coeff[0, H2Oi]) # with Raoult term
			
			
			dydt = y[H2Oi]-Csit[H2Oi] # distance from equilibrium (molecules/cc)

			dydt = ((dydt/Psat[0, H2Oi])*Cw)/2.0

			# new estimate of concentration of condensed water (molecules/cc (air))
			y[num_speci*(num_sb)+H2Oi] += dydt
			
	print('Finished initiating water condensation')
	
	return(y, Varr, x, N_perbin, Vbou, rbou)



'''preparing the inputs for the ode solver Jacobian'''
# called once the gas-phase and particle-phase equations have been interrogated

# required modules
import numpy as np

def jac_setup(jac_den_indx, njac, comp_num, num_sb, num_eqn, nreac_g, nprod_g, rindx_g, pindx_g, jac_indx_g, wall_on, nreac_aq, nprod_aq, rindx_aq, pindx_aq, jac_indx_aq, num_asb):

	# inputs ---------------------------------------------
	# jac_den_indx - index of denominators for jacobian
	# njac - number of jacobian elements affected per equation
	# comp_num - number of components
	# num_sb - number of size bins (including any wall)
	# num_eqn - number of chemical reactions
	# nreac_g - number of reactants per reaction
	# nprod_g - number of products per reaction
	# rindx_g - index of reactants per equation
	# pindx_g - index of products per equation
	# jac_indx_g - index of Jacobian affected per equation
	# wall_on - flag for whether wall on or off
	# nreac_aq - number of reactants per equation
	# nprod_aq - number of products per equation
	# rindx_aq - index of reactants per equation
	# pindx_aq - index of prodcuts per equation
	# jac_indx_aq - index of jacobian affected per equation
	# num_asb - number of actual particle size bins (excluding wall) 
	# ----------------------------------------------------

	print('Preparing Jacobian inputs')

	# rows in Jacobian affected by this equation		 	
	rowvals = np.empty((0))
	# indices of rowvals representing each component being 
	# differentiated by in Jacobian, note, needs to start
	# as zero to represent first index of rowval
	colptrs = np.zeros((1))

	# ensure integer type
	jac_den_indx = jac_den_indx.astype(int)
	njac = njac.astype(int)
	
	# track columns of Jacobian affected
	col_tr = 0
	# length of Jacobian; add 1 to account for the gas phase
	len_jac = comp_num*(num_sb+1)

	# index for Jacobian - note can only be done after all equations checked as only then
	# is the total number of unique components known
	for eqni in range(num_eqn[0]): # equation loop
	
		# total number of components in this equation
		tot_comp = nreac_g[eqni]+nprod_g[eqni]
		# combined index of reactants and products in this equation
		totindx = np.append(rindx_g[eqni, 0:nreac_g[eqni]], pindx_g[eqni, 0:nprod_g[eqni]])
		
		# reactant loop (equivalent to columns of 2D Jacobian)
		for i in range(nreac_g[eqni]):
			# index of flattened Jacobian being affected in 
			# this equation, note this doesn't need to 
			# account for water and core component as below (when unique elements found) all
			# full Jacobian matrix elements that aren't indexed are
			# omitted to create the sparse Jacobian matrix  
			jac_indx_g[eqni, i*tot_comp:(i+1)*tot_comp] = rindx_g[eqni, i]*len_jac+(totindx)
			# check if rowvals already sufficiently long to contain this reactant  
			if rindx_g[eqni, i]<=(len(colptrs)-2):

				# start index of rowvals where it is represented
				r_st = int(colptrs[rindx_g[eqni, i]])
				# end index of rowvals where it is represented
				r_en = int(colptrs[rindx_g[eqni, i]+1])
				# check if the components affected have been 
				# represented in previous equations
				for i2 in totindx:
					# if this reactant has affected this component in 
					# a previous reaction then jac_indx_g (set above) will
					# sum its effect and no need to account for
					# the Jacobian index with rowvals and 
					# colptrs				
					if (any(rowvals[r_st:r_en] == i2)):
						continue
					else: # account for in Jacobian
						new_el = np.array((float(i2))).reshape(1)
						# index of where new index enters rowvals (to ensure
						# indices per Jacobian column are present in
						# ascending order)
						r_in = r_st+int(sum(rowvals[r_st:r_en]<new_el))
						rowvals = np.concatenate([rowvals[0:r_in], new_el, rowvals[r_in::]])
						colptrs[rindx_g[eqni, i]+1::] += 1
						r_en += 1
						
						
			else: # if not accounted for in previous reaction
				# rows in Jacobian affected by this reactant 
				# in this equation	 	
				rowvals = np.append(rowvals, np.unique(totindx))
				while (rindx_g[eqni, i]>col_tr):
					colptrs = np.append(colptrs, colptrs[-1])
					col_tr += 1
				# indices of rowvals
				colptrs = np.append(colptrs, colptrs[-1]+len(np.unique(totindx)))
				# track number of columns of Jacobian affected				
				col_tr += 1

	# because above jac_indx_g contains the indices of data assuming
	# a full, rather than sparse matrix, now correct to affect
	# the sparse matrix by reducing indices where gaps occur in
	# the complete data matrix
	# get unique values in jac_indx_g
	uni = np.unique(jac_indx_g)
	# now reduce these indices based on the number of indices
	# omitted
	diff = 0 # prepare for number of empty points
	for i in range(1, len(uni)):
		diff += (uni[i]-uni[i-1])-1
		jac_indx_g[jac_indx_g==uni[i]] -= diff
	
	# colptrs responds to appendages to rowvals above and therefore may not
	# include representatives for all components across all size bins and 
	# wall, therefore extend to
	# include all components, including water and seed material.
	# Remember that any final element of colptrs should represent the final
	# rowval index, so colptrs needs to have a length one greater than the
	# number of components+number of components*number of size bins
	col_shrt =  ((comp_num+2)+(comp_num+2)*num_sb+1)-len(colptrs)
	# rowvals indices for final columns of Jacobian		
	colptrs = np.append(colptrs, np.array((colptrs[-1])).repeat(col_shrt))
	
	if (num_eqn[1] > 0): # if aqueous-phase reactions present

		# aqueous-phase reaction part -----------------------------------------------------
		# container for Jacobian index affected by aqueous-phase reactions	
		jac_indx_aq = np.zeros((num_eqn[1], max(nreac_aq*(nreac_aq+nprod_aq))))
		# index of unique rows affected per reactant, if zero acts as a filler this omitted below
		col_tracker = np.unique(rindx_aq)
		# track rows affected per column
		row_tracker = np.ones((max(nreac_aq+nprod_aq), len(col_tracker)))*-1.
		# note that size bins are dealt with below loops
		for eqni in range(num_eqn[1]): # loop through reactions
			# total number of components affected per reactant
			tot_affcomp = nreac_aq[eqni]+nprod_aq[eqni]
			# combined index of reactants and products in this equation
			totindx = np.append(rindx_aq[eqni, 0:nreac_aq[eqni]], pindx_aq[eqni, 0:nprod_aq[eqni]])
			for ir in range(nreac_aq[eqni]):# loop through reactants (would be columns in 2D Jacobian)
				# number of full Jacobian elements passed before reaching this reactant, 
				# accounting for water and core
				st_indx = (comp_num+2)*(num_sb+1)*rindx_aq[eqni, ir]
				# flat Jacobian index affected by this equation, this then reduced to sparse 
				# matrix index below
				jac_indx_aq[eqni, ir*tot_affcomp:(ir+1)*tot_affcomp] = st_indx+totindx
			
				# the column number affected
				coli = np.where(col_tracker == rindx_aq[eqni, ir])[0]
				# unique rows affected per reactant
				unirow = np.unique(np.append(row_tracker[:, coli][row_tracker[:, coli]!=-1], totindx))
				row_tracker[0:len(unirow), coli] = unirow.reshape(-1, 1)
		
		# account for the the gas-phase Jacobian indices
		row_tracker[row_tracker != -1] += comp_num+2
		# flatten ready for appendage
		row_trackerf = row_tracker.flatten(order='F')
		# append to rowvals
		rowvals = np.append(rowvals, row_trackerf[row_trackerf != -1])
		# sum number of unique rows affected per component
		col_num = (row_tracker != -1).sum(axis = 0)
		# account for new rows in first size bin, add 1 to index because this represents the
		# number of elements per component
		colptrs[col_tracker+(comp_num+2)+1] += np.cumsum(col_num)
		# ensure latter inidices are consistent 
		colptrs[max(col_tracker+(comp_num+2)+1)::] = colptrs[max(col_tracker+(comp_num+2)+1)] 
		
		# because above jac_indx_aq contains the indices of data assuming
		# a full, rather than sparse matrix, now correct to affect
		# the sparse matrix by reducing indices where gaps occur in
		# the complete data matrix
		# get unique values, mask out elements equal to zero as these are 
		# unfilled from the original container
		uni = np.unique(jac_indx_aq[jac_indx_aq != 0.])
		# now reduce these unique indices based on the number of indices
		# omitted
		diff = 0 # prepare for number of empty points
		for i in range(1, len(uni)):
			diff += (uni[i]-uni[i-1])-1
			jac_indx_aq[jac_indx_aq==uni[i]] -= diff
		
		# now account for omission of full Jacobian elements between the gas-phase 
		# reaction part and the aqueous-phase reaction part
		jac_indx_aq[jac_indx_aq != 0.] -= np.amin(jac_indx_aq[jac_indx_aq != 0.])-np.amax(jac_indx_g)-1
		# range of aqueous-phase sparse jacobian indices
		ran_aq = np.amax(jac_indx_aq[jac_indx_aq != 0.])-np.amin(jac_indx_aq[jac_indx_aq != 0.])+1
		# shape of index matrix for just one size bin
		jsh = jac_indx_aq.shape
		# now repeat over particle size bins
		for sbi in range(1, num_asb):
			jac_indx_aq = np.concatenate((jac_indx_aq, jac_indx_aq[0:jsh[0], 0:jsh[1]]+ran_aq*(sbi)), axis = 0)
			
			# account for the the gas-phase Jacobian indices
			row_tracker[row_tracker != -1] += (comp_num+2)
			# flatten ready for appendage
			row_trackerf = row_tracker.flatten(order='F')
			# append to rowvals
			rowvals = np.append(rowvals, row_trackerf[row_trackerf != -1])
			# account for new rows in latter size bin, add 1 to index because this represents the
			# number of elements per component
			colptrs[col_tracker+(comp_num+2)*(sbi+1)+1] += np.cumsum(col_num)
			# ensure latter inidices are consistent 
			colptrs[max(col_tracker+(comp_num+2)*(sbi+1)+1)::] = colptrs[max(col_tracker+(comp_num+2)*(sbi+1)+1)]
	
	if (num_eqn[1] == 0):
		jac_indx_aq = np.zeros((1)) # filler

	# particle influence on Jacobian part -------------------------------------------------
	# loop through jacobian index to check whether the centre 
	# diagonal for gas effect on gas components is due to be 
	# affected by gas-phase reactions or whether particle 
	# partitioning is the first effect.  Similarly accommodate 
	# for particle effect for: gas on particle effect, particle on gas effect
	# and particle on particle effect

	# Jacobian index for particle effects
	jac_part_indx = np.zeros((comp_num+2)*(num_asb+1)+((comp_num+2)*(num_asb*2)))
	
	part_cnt = 0 # count on jac_part_indx inputs
	
	if (num_asb>0): # if particle size bins are present

		# loop through components in the gas-phase (add two 
		# to account for water and seed material)
		for compi in range(comp_num+2):
			# gas effect on gas part ---------------------------------------------
			# relevant starting and finishing index in rowvals
			st_indx = int(colptrs[compi])
			en_indx = int(colptrs[compi+1])
			if ((st_indx < en_indx) == True): # rows already attributed to this column
				# if rows are already attributed, check 
				# whether the diagonal is already affected
				if (any(rowvals[st_indx:en_indx] == compi)):
					# get index
					exist_indx = st_indx+(np.where(rowvals[st_indx:en_indx] == compi)[0][0])
					# if diagonal already affected then just copy relevant
					# data index to the indexing for wall
					jac_part_indx[part_cnt] = exist_indx
					
				else: # if diagonal not already affected, then include
					# get index
					new_indx = st_indx+(sum(rowvals[st_indx:en_indx]<compi))+1
					# modify indices for sparse Jacobian matrix
					jac_part_indx[part_cnt] = new_indx # gas-particle partitioning
					jac_indx_g[jac_indx_g>=new_indx] += 1 # gas-phase reactions
					jac_indx_aq[jac_indx_aq>=new_indx] += 1 # aqueous-phase reactions
					new_el = np.array((compi)).reshape(1) 
					rowvals = np.concatenate([rowvals[0:new_indx], 
						new_el, rowvals[new_indx::]])
					colptrs[compi+1::] += 1
					# increase en_indx to account for new row when
					# accommodating gas effect on wall below
					en_indx += 1
					
			else: # no rows yet attributed to this column, so need to include

				# modify indices for sparse Jacobian matrix
				jac_part_indx[part_cnt] = st_indx # gas-particle partitioning
				jac_indx_g[jac_indx_g>=st_indx] += 1 # gas-phase reactions
				jac_indx_aq[jac_indx_aq>=st_indx] += 1 # aqueous-phase reactions
				new_el = np.array((compi)).reshape(1)
				rowvals = np.concatenate([rowvals[0:st_indx], new_el, rowvals[st_indx::]])
				colptrs[compi+1::] += 1
				# increase en_indx to account for new row when accommodating
				# gas effect on particle below
				en_indx += 1
				
			part_cnt += 1 # keep count on particle index
			
			# gas effect on particle part -------------------------------
			# already know that the gas-phase reactions will not
			# have affected this part of the Jacobian, so just need to
			# modify sparse Jacobian inputs accordingly
			jac_part_indx[part_cnt:part_cnt+num_asb] = range(en_indx, en_indx+num_asb)
			jac_indx_g[jac_indx_g>=en_indx] += num_asb # gas-phase reaction
			jac_indx_aq[jac_indx_aq>=en_indx] += num_asb # aqueous-phase reactions
			new_el = (np.array(range(comp_num+2+compi, 
				(comp_num+2)*(num_asb+1)+compi, (comp_num+2)))).reshape(-1)
			rowvals = np.concatenate([rowvals[0:en_indx], new_el, rowvals[en_indx::]])
			colptrs[compi+1::] += num_asb
			part_cnt += num_asb # keep count on particle index		
		
		
		# loop through components to include their particle effect on
		# gas phase and on particle components for the Jacobian, note if no wall this 
		# will include the final row in the final column of the Jacobian
		for sbi in range(num_asb): # size bin loop
			for compi in range(comp_num+2): # component loop
				# particle component effect on gas-phase
				
				# the flattened Jacobian index relating to this
				# relevant start index for colptrs
				stc_indx = int((comp_num+2)*(sbi+1)+(compi))
				st_indx = int(colptrs[stc_indx]) # start index for rowvals		
				jac_part_indx[part_cnt] = st_indx
				new_el = np.array((compi)).reshape(-1)
				rowvals = np.concatenate([rowvals[0:st_indx], new_el, rowvals[st_indx::]])
				colptrs[(comp_num+2)*(sbi+1)+(compi+1)::] += 1
				jac_indx_aq[jac_indx_aq>=st_indx] += 1 # aqueous-phase reactions
				part_cnt += 1
				
				# particle component effect on particle
				new_el = np.array(((comp_num+2)*(sbi+1)+compi)).reshape(-1)
				# account for any aqueous-phase reactions
				st_indx = (st_indx) + sum(rowvals[int(colptrs[stc_indx]):int(colptrs[stc_indx+1])] < new_el)
				jac_part_indx[part_cnt] = st_indx
				# check if diagonal already in use by aqueous-phase reactions
				if any(rowvals[int(colptrs[stc_indx]):int(colptrs[stc_indx+1])] == new_el):
					part_cnt += 1
					continue # continue to next component if it is
				else:
					rowvals = np.concatenate([rowvals[0:st_indx], new_el, rowvals[st_indx::]])
					colptrs[(comp_num+2)*(sbi+1)+compi+1::] += 1
					jac_indx_aq[jac_indx_aq>=st_indx] += 1 # aqueous-phase reactions
					part_cnt += 1
	
	# wall influence on Jacobian part ---------------------------------------------------

	# loop through jacobian index to check whether the centre 
	# diagonal for gas effect on gas components is due to be 
	# affected by gas-phase reactions or whether wall 
	# partitioning is the first effect.  Similarly accommodate 
	# for wall effect for: gas on wall effect, wall on gas effect
	# and wall on wall affect
	jac_wall_indx = np.zeros(((comp_num+2)*4))
	wall_cnt = 0 # count on jac_wall_indx inputs

	if (wall_on > 0):
		# loop through components in the gas-phase (add two 
		# to account for water and core component)
		for compi in range(comp_num+2):
			# gas effect on gas part --------------------------------------------
			# relevant starting and finishing index in rowvals
			st_indx = int(colptrs[compi])
			en_indx = int(colptrs[compi+1])
			
			# check if any rows already attributed to this column
			if ((st_indx<en_indx) == True):
				# if rows are already attributed, check 
				# whether the diagonal is already affected
				if (sum(rowvals[st_indx:en_indx]==compi))>0:
					# get index
					exist_indx = st_indx+(np.where(rowvals[st_indx:en_indx]==compi)[0][0])
					# if diagonal already affected then just copy relevant
					# data index to the indexing for wall
					jac_wall_indx[wall_cnt] = exist_indx
					
				else: # if diagonal not already affected, then add
					# get index
					new_indx = st_indx+(sum(rowvals[st_indx:en_indx]<compi))+1
					# modify indices for sparse Jacobian matrix
					jac_wall_indx[wall_cnt] = new_indx
					jac_indx_g[jac_indx_g>=new_indx] += 1
					jac_indx_aq[jac_indx_aq>=new_indx] += 1
					new_el = np.array((compi)).reshape(1)
					rowvals = np.concatenate([rowvals[0:new_indx], 
						new_el, rowvals[new_indx::]])
					colptrs[compi+1::] += 1
					# increase en_indx to account for new row when
					# accommodating gas effect on wall below
					en_indx += 1
					
			else: # no rows yet attributed to this column, so need to include

				# modify indices for sparse Jacobian matrix
				jac_wall_indx[wall_cnt] = st_indx
				jac_indx_g[jac_indx_g>=st_indx] += 1
				jac_indx_aq[jac_indx_aq>=st_indx] += 1
				new_el = np.array((compi)).reshape(1)
				rowvals = np.concatenate([rowvals[0:st_indx], new_el, rowvals[st_indx::]])
				colptrs[compi+1::] += 1
				# increase en_indx to account for new row when accommodating
				# gas effect on wall below
				en_indx += 1
				
			wall_cnt += 1 # keep count on wall index
			
			# gas effect on wall part -----------------------------------------
			# already know that the gas-phase photochemistry will not
			# have affected this part of the Jacobian, so just need to
			# modify sparse Jacobian inputs accordingly
			jac_wall_indx[wall_cnt] = en_indx
			# bump gas effect and particle effect indices up by one to 
			# account for addition to Jacobian
			jac_indx_g[jac_indx_g>=en_indx] += 1
			jac_indx_aq[jac_indx_aq>=en_indx] += 1
			if (num_asb > 0):
				jac_part_indx[jac_part_indx>=en_indx] += 1
			new_el = np.array(((comp_num+2)*(num_asb+1)+compi)).reshape(1)
			rowvals = np.concatenate([rowvals[0:en_indx], new_el, rowvals[en_indx::]])
			colptrs[compi+1::] += 1
			wall_cnt += 1 # keep count on wall index
		
		# wall effect on gas phase and on wall components for the Jacobian, note this will include the
		# final row in the final column of the Jacobian
		for compi in range(comp_num+2):
			# wall component effect on itself in gas-phase
			jac_wall_indx[wall_cnt] = len(rowvals)
			new_el = np.array((compi)).reshape(1)
			rowvals = np.append(rowvals, new_el)
			colptrs[(comp_num+2)*(num_asb+1)+compi+1::] += 1
			wall_cnt += 1
			# wall component effect on itself on wall
			jac_wall_indx[wall_cnt] = len(rowvals)
			new_el = np.array(((comp_num+2)*(num_asb+1)+compi)).reshape(1)
			rowvals = np.append(rowvals, new_el)
			colptrs[(comp_num+2)*(num_asb+1)+compi+1::] += 1
			wall_cnt += 1
	# end of wall influence on Jacobian part ---------------------------------------------------
		 		
	if (num_sb == 0): # if no particle size bins and no wall

		# if the Jacobian matrix has an empty final row, then
		# an error will be displayed during ODE solver call, so 
		# input a filler on this row;
		# add two to the number of unique components counted here to
		# acount for water and seed material
		if max(rowvals)<=(comp_num+2)-1:
			# index for final row of Jacobian
			rowvals = np.append(rowvals, (comp_num+2)-1)
			# number of columns currently short of final column
			col_shrt =  comp_num+2-len(colptrs)
			# rowval index for final column of Jacobian		
			colptrs[-1] += 1

	# ensure integer type for arrays acting as indices
	jac_indx_g = jac_indx_g.astype(int)
	jac_indx_aq = jac_indx_aq.astype(int)
	jac_part_indx = jac_part_indx.astype(int)
	jac_wall_indx = jac_wall_indx.astype(int)	
	rowvals = rowvals.astype(int)	
	colptrs = colptrs.astype(int)



	return(rowvals, colptrs, jac_indx_g, jac_indx_aq, jac_part_indx, jac_wall_indx)



'''module to save PyCHAM results to files'''
# module called following simulation in PyCHAM to store generated and input information

import os
import sys
import numpy as np

def saving(filename, y_mat, Nresult_dry, Nresult_wet, t_out, savefolder, dydt_vst, num_comp, 
	Cfactor_vst, testf, numsb, comp_namelist, dydt_trak, y_mw, MV,
	time_taken, seed_name, x2, rbou_rec, wall_on, space_mode, rbou00, upper_bin_rad_amp):

	# inputs: ----------------------------------------------------------------------------
	
	# filename - name of inputs file name
	# y_mat - species (columns) concentrations with time (rows) (molecules/cc (air))
	# Nresult_dry  - number concentration of dry particles per size bin (#/cc (air))
	# Nresult_wet  - number concentration of dry particles per size bin (#/cc (air))
	# Cfactor - conversion factor to change gas-phase concentrations from molecules/cc 
	# (air) into ppb
	# testf - flag to show whether in normal mode (0) or test mode (1)
	# numsb - number of size bins
	# dydt_vst - tendency to change of user-specified components
	# dydt_trak - user-input names of components to track
	# comp_namelist - names of species given by the equation file
	# upper_bin_rad_amp - factor upper bin radius found increased by in 
	#						Size_distributions.py for more than 1 size bin, or in 
	# 						pp_intro.py for 1 size bin
	# Cfactor_vst - one billionth the molecular concentration in a unit volume of chamber
	#				(molecules/cc) per recording time step
	# time_taken - computer time for entire simulation (s)
	# seed_name - name of seed component
	# comp_namelist - chemical scheme names of components 
	# y_mw - molecular weights (g/mol)
	# MV - molar volumes (cm3/mol)
	# time_taken - simulation computer time (s)
	# seed_name - chemical scheme name of component comprising seed particles
	# x2 - record of size bin radii (um)
	# rbou_rec - radius bounds per size bin (columns) per time step (rows) (um)
	# wall_on - marker for whether wall on or off
	# space_mode - type of spacing between particle size bins (log or lin)
	# rbou00 - initial lower size (radius) bound of particles (um)
	# upper_bin_rad_amp - factor increase in radius of upper bound 
	# -------------------------------
	
	if (numsb-wall_on > 0): # correct for changes to size bin radius bounds
		rbou_rec[:, 0] = rbou00
		rbou_rec[:, -1] = rbou_rec[:, -1]/upper_bin_rad_amp

	if (testf == 1):
		return(0) # return dummy

	dir_path = os.getcwd() # current working directory
	output_root = 'PyCHAM/output'
	filename = os.path.basename(filename)
	filename = os.path.splitext(filename)[0]
	# one folder for one simulation
	output_by_sim = os.path.join(dir_path, output_root, filename, savefolder)
	# create folder to store results
	os.makedirs(output_by_sim)
	
	# saving dictionary
	# dictionary containing variables for model and components
	const = {}
	const["number_of_size_bins"] = numsb
	const["number_of_components"] = num_comp
	const["molecular_weights_g/mol_corresponding_to_component_names"] = (np.squeeze(y_mw[:, 0]).tolist())
	const["molar_volumes_cm3/mol"] = (MV[:, 0].tolist())
	const["component_names"] = comp_namelist
	const["factor_for_multiplying_ppb_to_get_molec/cm3_with_time"] = (Cfactor_vst.tolist())
	const["simulation_computer_time(s)"] = time_taken
	const["seed_name"] = seed_name
	const["wall_on_flag_0forNO_1forYES"] = wall_on
	const["space_mode"] = space_mode
	with open(os.path.join(output_by_sim,'model_and_component_constants'),'w') as f:
		for key in const.keys():
			f.write("%s,%s\n"%(key, const[key]))
	
	# convert gas-phase concentrations from molecules/cc (air) into ppb
	# leaving any particle-phase concentrations as molecules/cc (air)
	y_mat[:, 0:num_comp] = y_mat[:, 0:num_comp]/(Cfactor_vst.reshape(len(Cfactor_vst), 1))
	
	# prepare header for concentrations with time file 
	y_header = str('')
	x2_header = str('') # prepare header for files relating to size bins
	for i in range(numsb+1):

		if i == 0:
			end = '_g'
		if (i>0 and i<numsb):
			end = '_p'
		if i==numsb:
			if wall_on==0:
				end = '_p'
			else:
				end = '_w'
		for ii in range(num_comp):
			if i  == 0 and ii == 0:
				start = ''
			else:
				start = ', '
			y_header = str(y_header+str(start+comp_namelist[ii])+end)
			
	# saving both gas- and particle-phase concentrations of species
	np.savetxt(os.path.join(output_by_sim, 'concentrations_all_components_all_times_gas_particle_wall'), y_mat, delimiter=',', header=str('time changes with rows which correspond to the time output file, components in columns, with _g representing gas phase (ppb), _pi representing particle phase where i is the size bin number (starting at 1) (molecules/cc (air)) and _w is the wall phase (molecules/cc (air))\n'+y_header)) 	
	 	
	# saving time of outputs
	np.savetxt(os.path.join(output_by_sim, 'time'), t_out, delimiter=',', header='time (s), these correspond to the rows in the concentrations_all_components_all_times_gas_particle_wall, particle_number_concentration and size_bin_radius output files') 	
	
	# if tracking of tendencies to change requested by user, loop through the components
	# and save the tendency record for each of these (%/hour)
	if len(dydt_vst)>0:
		compind = 0
		# loop through components to record the tendency of change \n')
		for compi in dydt_vst.get('comp_index'):
			# open relevant dictionary value, to get the 2D numpy array for saving
			dydt_rec = np.array(dydt_vst.get(compi))
			
			# get user-input name of this component
			comp_name = str(dydt_trak[compind] +'_rate_of_change')
			# save
			np.savetxt(os.path.join(output_by_sim, comp_name), dydt_rec, delimiter=',', header='tendency to change, top row gives equation number (where number 0 is the first equation) (molecules/cc.s (air))')
			compind += 1
	
	
	if (numsb-wall_on > 0): # if particles present
	
		np.savetxt(os.path.join(output_by_sim, 'particle_number_concentration_dry'), Nresult_dry, delimiter=',',
				header=('particle number concentration assuming water removed from particles (#/cc (air)), with time changing with rows (corresponding times given in the time output file) and size bin changing with columns with size bin numbers given in the second row of the header\n'+x2_header))
		
		np.savetxt(os.path.join(output_by_sim, 'particle_number_concentration_wet'), Nresult_wet, delimiter=',',
				header=('particle number concentration assuming water not removed from particles (#/cc (air)), with time changing with rows (corresponding times given in the time output file) and size bin changing with columns with size bin numbers given in the second row of the header\n'+x2_header))	
	
		np.savetxt(os.path.join(output_by_sim, 'size_bin_radius'), x2, delimiter=',',
				header= str('particle radii (um) per size_bin (including water contribution to size), with size bins represented by columns and their number (starting from 1) given in second line of header, per time step which is represented by rows and corresponding times given in the time output file \n'+x2_header))
	
		np.savetxt(os.path.join(output_by_sim, 'size_bin_bounds'), rbou_rec, delimiter=',',
				header=str('particle size bin bounds (um), with size bins represented by columns and their number (starting at 1 and in line with the lower bound) given in second line of header, per time step which is is represented by rows and corresponding times given in the time output file \n'+x2_header))		
		
	# if save name is the default, then remove to ensure no duplication in future
	if (savefolder == 'default_res_name'):
		import shutil
		print('Note: default name for save folder used, therefore now deleting this folder to avoid future duplication')	
		shutil.rmtree(output_by_sim)
	return()



'''module to implement nucleation in PyCHAM'''

import numpy as np
import scipy.constants as si

def nuc(sumt, new_part_sum1, n0, y, MW, rho, num_comp, Varr, x, new_partr, MV, 
		nucv1, nucv2, nucv3, nuc_comp, siz_str, rbou, Vbou, num_sb):


	# ---------------------------------------------------------------
	# input:
	
	# sumt - time through simulation (s)
	# new_part_sum1 - number concentration of newly nucleated particles already 
	# added (# particles/cc (air))
	# n0 - original number concentration of particles (# particles/cc (air))
	# y - molecular concentration of components in the gas and particle phase
	# (molecules/cc (air))
	# MW - molecular weight (g/mol)
	# rho - density (g/cc)
	# num_comp - number of components
	# Varr - particles volumes per size bin(s) (um3)	
	# x - particles radius per size bin(s) (um)
	# new_partr - radius of newly nucleated particle (cm)
	# MV - molar volume (cc/mol)
	# nucv1/v2/v3 - parameter values for nucleation equation
	# nuc_comp - index of the nucleating component
	# siz_str - the size structure
	# rbou - radius bounds (um)
	# Vbou - volume bounds (um3)
	# num_sb - number of size bins, excluding wall
	# ---------------------------------------------------------------
	
	# total number of particles that should have been produced through nucleation 
	# by now given by integration:

	# Gompertz function (for asymmetric sigmoid curve): https://en.wikipedia.org/wiki/Gompertz_function
	nsum1 = nucv1*np.exp(nucv2*(np.exp(-sumt/nucv3)))

	# new number of particles this time step
	new_part1 = (nsum1)-new_part_sum1
	
	# note, setting condition to less than zero may cause very small concentrations of particles 
	# to appear very far from the peak nucleation time
	if (new_part1 < 1.):
		new_part1 = 0.
		
	else: # if new particles are being added
		if (siz_str == 1): # if full-moving being used
			# if there is geometric space for newly formed particles to form a new size bin
			if (rbou[1]/2. > new_partr*1.e4): # scale new_partr to convert from cm to um
				# form new size bin for newly formed particles
				n0[-2] += n0[-1]
				n0 = np.concatenate((np.zeros((1, 1)), n0[0:-1]), axis = 0)
				y[-num_comp*3:-num_comp*2] += y[-num_comp*2:-num_comp*1]
				# remove final size bin
				y = np.concatenate((y[0:-num_comp*2], y[-num_comp::]))
				rbou = np.concatenate((rbou[0:-2], rbou[-1].reshape(1)))
				# add new size bin
				y = np.concatenate((y[0:num_comp], np.zeros((num_comp)), y[num_comp:]))
				rbou = np.concatenate((rbou[0].reshape(1), (rbou[1]/2.).reshape(1), rbou[1::]))
				Vbou = (4./3.)*np.pi*rbou**3.
				
	# identify size bin newly formed particles should enter
	sbi = sum(new_partr*1.e4 > rbou[1::]) # scale new_partr to convert from cm to um
	
	n0[sbi] += new_part1 # add to this size bin
	new_part_sum1 += new_part1
	
	# volume concentration of new particles (cc/cc (air))
	new_vol1 = new_part1*((4./3.)*np.pi*(new_partr)**3.)
	
	# molecular volume of nucleating component
	Vpermolec = (MV[nuc_comp])/si.N_A # molecular volume (cc/molecule)
	# concentration of nucleating component this represents (molecules/cc (air))
	nuc_conc1 = new_vol1/Vpermolec
	
	# remove from gas-phase (molecules/cc (air)), note commented out as violation of
	# mass conservation assumed negligible
# 	y[nuc_comp] -= nuc_conc1
		
	# addition to particle-phase (molecules/cc (air))
	y[num_comp*(1+sbi)+nuc_comp] += nuc_conc1
	
	# average volume of single particles now (scale MV by 1.e12 to convert 
	# from cm3 to um3), note, consider all size bins since if full-moving used,
	# the smallest and largest size bin will be affected
	ypsb = y[num_comp:num_comp*(num_sb+1)].reshape(num_sb, num_comp, order = 'C')
	
	ish = (n0[:, 0] > 0) # index of size bins containing particles
	if (sum(ish) > 0): # if any bins contain particles
		Varr[ish] = ((ypsb[ish, :]/(si.N_A*n0[ish, :]))*(MV[:].reshape(1, -1)*1.e12)).sum(axis = 1)
		# average radius of particles (um)
		x = ((3.*Varr)/(4.*np.pi))**(1./3.)
		x = x.reshape(-1) # ensure x remains as 1D array
	
	return(n0, y, x, Varr, new_part_sum1, rbou, Vbou)



'''parses the input files to automatically create the solver file'''
# input files are interpreted and used to create the necessary
# arrays and python files to solve problem

import numpy as np
import sch_interr
import xml_interr
import eqn_interr
import photo_num
import write_RO2_indices
import write_dydt_rec
import write_ode_solv
import write_rate_file
import jac_setup
import aq_mat_prep

# define function to extract the chemical mechanism
def extr_mech(sch_name, chem_sch_mrk, xml_name, photo_path, 
		con_infl_nam, int_tol, wall_on, num_sb, const_comp):

	# inputs: ----------------------------------------------------
	# sch_name - file name of chemical scheme
	# chem_sch_mrk - markers to identify different sections of 
	# 			the chemical scheme
	# xml_name - name of xml file
	# photo_path - path to file containing absorption 
	# 	cross-sections and quantum yields
	# con_infl_nam - chemical scheme names of components with 
	# 		constant influx
	# int_tol - integration tolerances
	# wall_on - marker for whether to include wall partitioning
	# num_sb - number of size bins (including any wall)
	# const_comp - chemical scheme name of components with 
	#	constant concentration
	# ------------------------------------------------------------

	print('Starting equation parsing')
	f_open_eqn = open(sch_name, mode='r') # open the chemical scheme file
	
	# read the file and store everything into a list
	total_list_eqn = f_open_eqn.readlines()
	f_open_eqn.close() # close file
	
	# interrogate scheme to list equations
	[eqn_list, aqeqn_list, eqn_num, rrc, rrc_name, 
		RO2_names] = sch_interr.sch_interr(total_list_eqn, chem_sch_mrk)
	
	# interrogate xml to list component names and SMILES
	[comp_smil, comp_name] = xml_interr.xml_interr(xml_name)

	# get equation information for chemical reactions
	[rindx_g, rstoi_g, pindx_g, pstoi_g, reac_coef_g, 
		nreac_g, nprod_g, jac_stoi_g, 
		jac_den_indx_g, njac_g, jac_indx_g, 				
		y_arr_g, y_rind_g, uni_y_rind_g, y_pind_g, 
		uni_y_pind_g, reac_col_g, prod_col_g, rstoi_flat_g, pstoi_flat_g, 
		rr_arr_g, rr_arr_p_g, rindx_aq, rstoi_aq, pindx_aq, pstoi_aq, reac_coef_aq, 
		nreac_aq, nprod_aq, jac_stoi_aq, 
		jac_den_indx_aq, njac_aq, jac_indx_aq, 				
		y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, 
		uni_y_pind_aq, reac_col_aq, prod_col_aq, rstoi_flat_aq, pstoi_flat_aq, 
		rr_arr_aq, rr_arr_p_aq, comp_namelist, comp_list, Pybel_objects, 
		comp_num] = eqn_interr.eqn_interr(eqn_num, 
		eqn_list, aqeqn_list, chem_sch_mrk, comp_name, comp_smil, num_sb, wall_on)
	
	print(('Number of gas-phase chemical reactions identified: ' + str(eqn_num[0])))
	print(('Number of unique components identified in chemical scheme file: ' + str(comp_num)))

	[rowvals, colptrs, jac_indx_g, jac_indx_aq, jac_part_indx, jac_wall_indx] = jac_setup.jac_setup(jac_den_indx_g, njac_g, comp_num, num_sb, eqn_num, nreac_g, nprod_g, rindx_g, pindx_g, jac_indx_g, wall_on, nreac_aq, nprod_aq, rindx_aq, pindx_aq, jac_indx_aq, (num_sb-wall_on))
	
	# prepare aqueous-phase reaction matrices for applying to reaction rate calculation
	if (eqn_num[1] > 0): # if aqueous-phase reactions present
		[rindx_aq, rstoi_aq, pindx_aq, pstoi_aq, reac_coef_aq, 
			nprod_aq, jac_stoi_aq, njac_aq,
			jac_den_indx_aq, jac_indx_aq, 				
			y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, 
			uni_y_pind_aq, reac_col_aq, prod_col_aq, rstoi_flat_aq, pstoi_flat_aq, 
			rr_arr_aq, rr_arr_p_aq] = aq_mat_prep.aq_mat_prep(rindx_aq, rstoi_aq, 
			pindx_aq, pstoi_aq, reac_coef_aq, 
			nprod_aq, jac_stoi_aq, njac_aq, 
			jac_den_indx_aq, jac_indx_aq, 				
			y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, 
			uni_y_pind_aq, reac_col_aq, prod_col_aq, rstoi_flat_aq, pstoi_flat_aq, 
			rr_arr_aq, rr_arr_p_aq, num_sb, wall_on, eqn_num[1], comp_num) 
	
	
	print('Generating modules dependent on chemical scheme')
	# get index of components with constant influx/concentration -----------
	# empty array for storing index of components with constant influx
	con_infl_indx = np.zeros((len(con_infl_nam)))
	con_C_indx = np.zeros((len(const_comp)))
	for i in range (len(con_infl_nam)):
		# index of where constant components occur in list of components
		con_infl_indx[i] = comp_namelist.index(con_infl_nam[i])

	for i in range (len(const_comp)):
		# index of where constant concnetration components occur in list 
		# of components
		con_C_indx[i] = comp_namelist.index(const_comp[i])
	# ---------------------------------------------------------------------

	# call function to generate ordinary differential equation (ODE)
	# solver module, add two to comp_num to account for water and core component
	write_ode_solv.ode_gen(con_infl_indx, int_tol, rowvals, wall_on, comp_num+2, 
			(num_sb-wall_on), con_C_indx, 0, eqn_num)
	
	# get index of components in the peroxy radical list
	RO2_indx = write_RO2_indices.write_RO2_indices(comp_namelist, RO2_names)

	# call function to generate reaction rate calculation module
	write_rate_file.write_rate_file(reac_coef_g, reac_coef_aq, rrc, rrc_name, 0)

	# call function to generate module that tracks change tendencies
	# of certain components
	write_dydt_rec.write_dydt_rec()

	# get number of photolysis equations
	Jlen = photo_num.photo_num(photo_path)

	return(rindx_g, pindx_g, rstoi_g, pstoi_g, nreac_g, nprod_g, jac_stoi_g, 
		njac_g, jac_den_indx_g, jac_indx_g, y_arr_g, y_rind_g,
		uni_y_rind_g, y_pind_g, uni_y_pind_g, reac_col_g, prod_col_g, 
		rstoi_flat_g, pstoi_flat_g, rr_arr_g, rr_arr_p_g, rowvals, colptrs, 
		jac_wall_indx, jac_part_indx, comp_num, RO2_indx, comp_list, 
		Pybel_objects, eqn_num, comp_namelist, Jlen, 
		rindx_aq, rstoi_aq, pindx_aq, pstoi_aq, reac_coef_aq, 
		nreac_aq, nprod_aq, jac_stoi_aq, 
		jac_den_indx_aq, njac_aq, jac_indx_aq, 				
		y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, 
		uni_y_pind_aq, reac_col_aq, prod_col_aq, rstoi_flat_aq, pstoi_flat_aq, 
		rr_arr_aq, rr_arr_p_aq, comp_name, comp_smil)



'''module to estimate the particle and wall partitioning coefficient'''
# the kimt_calc module is called at the start of the model loop time interval to update
# the mass transfer coefficient of gases to particles

import numpy as np
from part_prop import part_prop
import scipy.constants as si

def kimt_calc(y, mfp, num_sb, num_comp, accom_coeff, y_mw, surfT, R_gas, TEMP, NA, 
		y_dens, N_perbin, radius, Psat, therm_sp,
		H2Oi, act_coeff, wall_on, caller, partit_cutoff, Press, DStar_org):
	
	# inputs:---------------------------------------------------------------------------
	
	# y - concentration of components' molecules (molecules/cc (air))
	# mfp - mean free path of gas molecules (m) (num_comp, 1)
	# accom_coeff - accommodation coefficients of components in each size bin
	# y_mw - molecular weight of components (g/mol)
	# radius - particle radius (m)
	# surfT - surface tension (g/s2==mN/m==dyn/cm)
	# R_gas - the universal gas constant (cc.Pa/K.mol)
	# TEMP - current temperature in chamber (K)
	# NA - Avogadro's constant (molecules/mol) 
	# y_dens - density of components (kg/m3)
	# N_perbin - number of particles per size bin (excluding wall)
	# Psat - liquid-phase saturation vapour pressures of components (molecules/cc (air))	
	# therm_sp - thermal speed of components (m/s) (num_comp, 1)
	# H2Oi - water index (integer)
	# act_coeff - activity coefficient of components (dimensionless)
	# wall_on - marker for whether wall present
	# caller - marker for the calling function
	# partit_cutoff - the product of Psat and act_coeff above which gas-particle 
	# 		partitioning assumed zero (Pa)
	# Press - air pressure (Pa)
	# DStar_org - gas-phase diffusion coefficient of components (cm2/s)
	# ------------------------------------------------------------------------------------
	
	if num_sb == 0: # fillers
		kimt = 0.
		kelv = 0.

		return(kimt, kelv)

	if (num_sb > 0) and (wall_on > 0): # if wall present
		y_part = y[num_comp:-(num_comp)]
	if (num_sb > 0) and (wall_on == 0): # if wall absent
		y_part = y[num_comp::]
	
	# density (g/cm3) and average molecular weight (g/mol) of particles (excluding wall)
	[tot_rho, ish, avMW] = part_prop(y_part, num_comp, (num_sb-wall_on), NA, y_mw, y_dens, 
					N_perbin)
	
	# Knudsen number (dimensionless)
	Kn = np.repeat(mfp, (num_sb-wall_on), 1)/np.repeat(radius, num_comp, 0)
	
	# update accommodation coefficients if necessary
	# note, using __import__ rather than import allows opening in run time, thereby using
	# updated module
	accom_coeff_calc = __import__('accom_coeff_calc')
	accom_coeff_now = accom_coeff_calc.accom_coeff_func(accom_coeff, radius)

	# Non-continuum regime correction 
	# calculate a correction factor according to the continuum versus non-continuum 
	# regimes
	# expression taken from Jacobson et al (2000), page 457, or Jacobson (2005), page 530 
	# (eq. 16.19).
	# They reference:
	# Fuchs and Sutugin 1971
	# Pruppacher and Klett 1997
	Inverse_Kn = Kn**-1.
	correct_1 = (1.33+0.71*Inverse_Kn)/(1.+Inverse_Kn)
	correct_2 = (4.*(1.-accom_coeff_now))/(3.*accom_coeff_now)
	correct_3 = 1.e0+(correct_1+correct_2)*Kn
	correction = correct_3**-1.
	
	# kelvin factor for each size bin (excluding wall), eq. 16.33 Jacobson et al. (2005)
	# note that avMW has units g/mol, surfT (g/s2==mN/m==dyn/cm), R_gas is multiplied by 
	# 1e7 for units g cm2/s2.mol.K, 
	# TEMP is K, radius is multiplied by 1e2 to give cm and tot_rho is g/cm3
	kelv = np.zeros((num_sb-wall_on, 1))
	kelv[ish, 0] = np.exp((2.e0*avMW[ish]*surfT)/(R_gas*1.e7*TEMP*radius[0, ish]*1.e2*tot_rho[ish]))


	# ------------------------------------------------
	# gas phase diffusion coefficient*Fuch-Sutugin correction (cm2/s)
	# eq. 5 Zaveri et al. (2008)
	kimt = (DStar_org)*correction
	# final partitioning coefficient (converting radius from m to cm)
	# eq. 16.2 of Jacobson (2005) and eq. 5 Zaveri et al. (2008)
	# components in rows and size bins in columns (/s)
	kimt = (4.*np.pi*(radius*1.e2)*N_perbin.reshape(1, -1))*kimt
	# transpose kimt ready for multiplication inside ode solver, so size bins
	# in rows and components in columns
	kimt = np.transpose(kimt)

	# zero partitioning to particles for any components with low condensability
	if (partit_cutoff): # if a value provided (default is empty list)

		# convert partit_cutoff from Pa to molecules/cc (air), note README states
		# that just one value accepted for partit_cutoff input
		partit_cutoff_Pa = partit_cutoff[0]*(NA/((R_gas*1.e6)*TEMP))
		highVPi = (Psat*act_coeff) > partit_cutoff_Pa
		highVPi[:, H2Oi] = 0 # mask water to allow its partitioning
		kimt[highVPi] = 0.

	# zero partitioning coefficient for size bins where no particles - enables significant
	# computation time acceleration and is physically realistic
	ish = N_perbin[:, 0]<=1.e-10
	kimt[ish, :] = 0.
	
	return(kimt, kelv)



'''module to calculate the integration needed for coagulation, called on by coag'''

import numpy as np
import scipy.constants as si
import scipy.integrate as integ
import pdb

def W_k_int(sbr_m, sbr_mt, sbr_sum_m, T, A_H, sbn):
	
	# -----------------------------------------------------------
	# inputs:
	
	# sbr_m - size bin radii spead across shells (m)
	# sbr_mt - transpose of sbr_m (m)
	# sbr_sum_m - sum of sbr_m and sbr_mt (m)
	# T - temperature (K)
	# A_H - Hamaker constant (dimensionless)
	# sbn - number of size bins
	# -----------------------------------------------------------
	
	# define the integrand in 15.43
	def integrand(x, a, b, c, m, A_H, rj, T):
	
		# differential of Ep with r=x/b substituted in prior to 
		# differentiation

		# van der Waals interaction potential (15.45)
		Ep0 = (-A_H/6.0)*((a)/((x/rj)**m-b)+
			(a)/((x/rj)**m-c)+
			np.log(((x/rj)**m-b)/((x/rj)**m-c)))
	
		# Ep with r=x/rj substituted in
		# first differential of Ep, all terms	
		Ep1 =  (-A_H/6.0)*(((-m*a*(x/(rj**m)))/(((x/rj)**m-b)**m)-
			(m*a*(x/(rj**m)))/(((x/rj)**m-c)**m))+
			((((m*x)/(rj**m))*((x**m)/(rj**m)-c)-
			((m*x)/(rj**m))*((x**m)/(rj**m)-b))/
			(((x**m)/(rj**m)-c)*((x**m)/(rj**m)-b))))	
	
		# second differential of third term
		f = (m*x*b)/(rj**m)-(m*x*c)/(rj**m)
		fd = (m*b)/(rj**m)-(m*c)/(rj**m)
		g = (x/rj)**4.0-((x/rj)**m)*b-((x/rj)**m)*c+c*b
		gd = (4.0*x**3.0)/(rj**4.0)-(m*x*b)/(rj**m)-(m*x*c)/(rj**m)
		
		# second differential of all terms
		Ep2 =  ((-A_H/6.0)*
			((((6.0*a*x**4.0)/(rj**6.0)-(4.0*x**m*a*b)/
			(rj**4.0)-(m*a*b**m)/(rj**m))/
			(((x/rj)**m-b)**4.0))+ 
			(((6.0*a*x**4.0)/(rj**6.0)-(4.0*x**m*a*c)/(rj**4.0)-
			(m*a*c**m)/(rj**m))/
			(((x/rj)**m-c)**4.0))+
			(fd*g-gd*f)/(g**m))) 
		# integration part of eq. 15.43
		return (((Ep1+(x/rj)*Ep2)*
			np.exp((-1.0/(si.k*T))*(((x/(rj*m))*Ep1+Ep0))))*
			((x/rj)**m))

	# define the integrand in 15.44 
	def integrand2(x, a, b, d, rj, e, m, c, T):
	
		# van der Waals interaction potential (15.45)
		Ep0 = (-A_H/6.0)*((m*a)/((x/rj)**m-b)+
			(m*a)/((x/rj)**m-c)+
			np.log(((x/rj)**m-b)/((x/rj)**m-c)))

		return (1.0+((2.6*a)/(b))*(((a)/(e*((x/rj)-d)))**0.5)+
		(a)/(e*((x/rj)-d)))*np.exp(Ep0/(si.k*T))*(1.0/((x/rj)**m))

	
	# numerator for first two terms in 15.45
	a = 2.0*(sbr_m*sbr_mt) 
	# square of sum of size bin radii (m2)
	b = sbr_sum_m**2.0
	# square of difference in size bin radii (m2)
	c = (sbr_m-sbr_mt)**2.0
	# difference in size bin radii (m2)
	d = sbr_m-sbr_mt
	# integration limits for 15.43
	ilu = 1.0+sbr_m/sbr_mt
	ill = 0.0	
	# constant for integration in eq. 15.43
	m = 2.0
	# empty matrices for the integration in 15.43 and the eq. 15.43 
	W_k_int = np.zeros((sbn, sbn))	
	W_k = np.zeros((sbn, sbn))

	# empty matrices for the integration in 15.44 and the eq. 15.44
	W_c_int = np.zeros((sbn, sbn))
	W_c = np.zeros((sbn, sbn))
		
	# loop through size bins
	for i in range(0, sbn):
		for j in range(i, sbn):	
			
			W_k_int[i,j] = integ.quad(integrand, ill, ilu[i,j], 
			args=(a[i,j], b[i,j], c[i,j], m, A_H, sbr_m[0,j], T))[0]
			W_k[i,j] = (-1.0/(m*b[i,j]*si.k*T))*W_k_int[i,j]

			W_c_int[i,j] = integ.quad(integrand2, ill, ilu[i,j], 
			args=(a[i,j]/2.0, b[i,j], d[i,j], sbr_m[0,j], 
			sbr_sum_m[i,j], m, c[i,j], T))[0]
			W_c[i,j] = 1.0/(sbr_sum_m[i,j]*W_c_int[i,j])

	return W_k, W_c




'''module to estimate component volatilities and liquid densities'''

# called/returned from/to the front.py and ode_gen.py modules, 
# this module is responsible for
# setting key properties of components, including liquid-phase saturation vapour pressures
# and liquid-phase densities.  It does this using either UManSysProp (default), or with
# user settings

import numpy as np
import sys
import os
from git import Repo
import shutil
import scipy.constants as si
import errno
import stat

def volat_calc(spec_list, Pybel_objects, TEMP, H2Oi, num_speci, Psat_water, vol_Comp, 
				volP, testf, corei, seed_name, pconc, umansysprop_update, core_dens, spec_namelist,
				ode_gen_flag, nuci, nuc_comp):

	# inputs: ------------------------------------------------------------
	# spec_list - array of SMILE strings for components 
	# (omitting water and core, if present)
	# Pybel_objects - list of Pybel objects representing the species in spec_list
	# (omitting water and core, if present)
	# TEMP - temperature (K) in chamber at time function called
	# vol_Comp - names of components (corresponding to those in chemical scheme file)
	# 			that have vapour pressures manually set in volP
	# testf - flag for whether in normal mode (0) or testing mode (1)
	# corei - index of seed particle component
	# seed_name - name(s) of components(s) comprising seed particles
	# pconc - initial number concentration of particles (#/cc (air))
	# umansysprop_update - marker for cloning UManSysProp so that latest version used
	# core_dens - density of core material (g/cc (liquid/solid density))
	# spec_namelist - list of components' names in chemical equation file
	# ode_gen_flag - whether or not called from front or ode_gen
	# nuci - index of nucleating component
	# nuc_comp - name of nucleating component
	# ------------------------------------------------------------
	
	
	
	if testf==1:
		return(0,0,0) # return dummies
		
	cwd = os.getcwd() # address of current working directory
	if umansysprop_update == 1:
		print('Cloning latest version of UManSysProp in volat_calc module')
		# download latest version of umansysprop
		
		# check if there is an existing umansysprop folder
		if os.path.isdir(cwd + '/umansysprop'): 
			def handleRemoveReadonly(func, path, exc):
				excvalue = exc[1]
				if not os.access(path, os.W_OK):
					# Is the error an access error ?
					os.chmod(path, stat.S_IWUSR)
					func(path)
				else:
					raise
			# remove existing folder, onerror will change permission of directory if 
			# needed
			shutil.rmtree(cwd + '/umansysprop', ignore_errors=False, onerror=handleRemoveReadonly)
		
		git_url = 'https://github.com/loftytopping/UManSysProp_public.git'
		Repo.clone_from(git_url, (cwd + '/umansysprop'))
	
	# point to umansysprop folder
	sys.path.insert(1, (cwd + '/umansysprop')) # address for updated version
	
	from umansysprop import boiling_points
	from umansysprop import vapour_pressures
	from umansysprop import liquid_densities

	NA = si.Avogadro # Avogadro's number (molecules/mol)
	y_dens = np.zeros((num_speci, 1)) # components' liquid density (kg/m3)
	Psat = np.zeros((num_speci, 1)) # species' vapour pressure

	
	if ode_gen_flag == 0: # estimate densities
		
		for i in range (num_speci):
			
			# density estimation ---------------------------------------------------------
			if i == H2Oi:
				y_dens[i] = 1.0*1.0E3 # (kg/m3 (particle))
				continue
			# core properties
			if (i == corei[0]):
				y_dens[i] = core_dens*1.e3 # core density (kg/m3 (particle))
				continue
			# nucleating component density, if component is core (kg/m3 (particle))
			if i == nuci and nuc_comp[0] == 'core': 
				y_dens[i] = 1.0*1.0E3
				continue
			if spec_list[i] == '[HH]': # omit H2 as unliked by liquid density code
				# liquid density code does not like H2, so manually input kg/m3
				y_dens[i] = 1.0e3
			else:
				# density (convert from g/cc to kg/m3)
				y_dens[i] = liquid_densities.girolami(Pybel_objects[i])*1.0E3
			# ----------------------------------------------------------------------------
	
	# estimate vapour pressures (log10(atm))
	for i in range (num_speci):
		
		if (i == corei[0]): # if this core component
			continue # core component not included in Pybel_objects
		if i == nuci and nuc_comp[0] == 'core':
			continue # core component not included in Pybel_objects
		
		# water vapour pressure already given by Psat_water (log10(atm))
		if i == H2Oi:
			Psat[i] = Psat_water
			continue # water not included in Pybel_objects
		
		# vapour pressure (log10 atm) (# eq. 6 of Nannoolal et al. (2008), with dB of 
		# that equation given by eq. 7 of same reference)
		Psat[i] = ((vapour_pressures.nannoolal(Pybel_objects[i], TEMP, 
						boiling_points.nannoolal(Pybel_objects[i]))))
	
	ish = Psat==0.0
	
	Psat = (np.power(10.0, Psat)*101325.0) # convert to Pa from atm
	# retain low volatility where wanted
	Psat[ish] = 0.0
	
	# manually assigned vapour pressures (Pa)
	if len(vol_Comp)>0 and ode_gen_flag==0:
		for i in range (len(vol_Comp)):
			# index of component in list of components
			vol_indx = spec_namelist.index(vol_Comp[i])
			Psat[vol_indx, 0] = volP[i]
	# ensure if nucleating component is core that it is involatile
	if nuc_comp == 'core':
		Psat[nuci, 0] = 0.0
	
	Psat_Pa = np.zeros((len(Psat), 1)) # for storing vapour pressures in Pa (Pa)
	Psat_Pa[:, 0] = Psat[:, 0]
    
	# convert saturation vapour pressures from Pa to molecules/cc (air) using ideal
	# gas law, R has units cc.Pa/K.mol
	Psat = Psat*(NA/((si.R*1.e6)*TEMP))
	
	return Psat, y_dens, Psat_Pa



'''module for calculating accommodation coefficients, automatically generated by kimt_prep'''

# code that expresses and performs the functions for accommodation 
# coefficients that are given by the user in the model variables file 
# and that are intended to be calculated real-time via the kimt_calc 
# module 

def accom_coeff_func(accom_coeff, radius):

	# ------------------------------------------------------------------ 
	# inputs:
	# accom_coeff - array containing accommdotion coefficients for all 
	# components
	# radius - radii of size bins (m)
	# ------------------------------------------------------------------ 

	# calculate accommodation coefficients 

	return(accom_coeff)



'''updates integration constants and calls ODE solver'''
# constants for the integration of the relevant ODEs are updated
# at intervals and passed to the ODE solver to obtain solutions

import numpy as np
import math as math
import rrc_calc # function to update rate coefficients
import cham_up
import rec_prep
import partit_var
import rec
import jac_up
import mov_cen
import fullmov
import wallloss
import nuc
import coag

def ode_updater(update_stp, 
	tot_time, save_stp, y, rindx, 
	pindx, rstoi, pstoi, nreac, nprod, jac_stoi, njac, 
	jac_den_indx, jac_indx, RO2_indx, H2Oi,	temp, tempt, 
	Pnow, light_stat, light_time, daytime, lat, lon, af_path, 
	dayOfYear, photo_path, Jlen, con_infl_C, nrec_steps, 
	dydt_vst, siz_str, num_sb, num_comp, seedi, seed_name, seedVr, 
	core_diss, Psat, mfp, therm_sp,
	accom_coeff, y_mw, surfT, R_gas, NA, y_dens, 
	x, Varr, act_coeff, Cw, kw, Cfactor, tf, light_ad, y_arr, 
	y_rind, 
	uni_y_rind, y_pind, uni_y_pind, reac_col, prod_col, 
	rstoi_flat, pstoi_flat, rr_arr, rr_arr_p, rowvals, 
	colptrs, wall_on, jac_wall_indx, jac_part_indx, Vbou,
	N_perbin, Vol0, rad0, np_sum, new_partr, nucv1, nucv2, 
	nucv3, nuc_comp, nuc_ad, RH, coag_on, inflectDp, pwl_xpre, 
	pwl_xpro, inflectk, chamR, McMurry_flag, p_char, e_field, 
	injectt, inj_indx, Ct, pmode, pconc, pconct, mean_rad, lowsize, 
	uppsize, std, rbou, const_infl_t, MV,
	rindx_aq, 
	pindx_aq, rstoi_aq, pstoi_aq, nreac_aq, nprod_aq, jac_stoi_aq, njac_aq, 
	jac_den_indx_aq, jac_indx_aq, y_arr_aq,
	y_rind_aq, 
	uni_y_rind_aq, y_pind_aq, uni_y_pind_aq, reac_col_aq, prod_col_aq, 
	rstoi_flat_aq, pstoi_flat_aq, rr_arr_aq, rr_arr_p_aq, eqn_num, 
	partit_cutoff, diff_vol, DStar_org, corei, ser_H2O):

	import ode_solv # import most updated version
	import ode_solv_wat # import most updated version
	
	# inputs: ----------------------------------------------------
	# update_stp - interval at which to update integration 
	#		constants (s)
	# tot_time - total time to simulate (s)
	# save_stp - frequency to save results (s)
	# y - initial component concentrations (molecules/cc/s)
	# rindx - index of reactants per equation
	# pindx - index of products per equation
	# nreac - number of reactants per equation
	# nprod - number of products per equation
	# jac_stoi - stoichiometries relevant to Jacobian
	# njac - number of elements of Jacobian affected per equation
	# jac_den_indx - index of denominator components for Jacobian
	# jac_indx - index of Jacobian affected per equation
	# RO2_indx - index of components in peroxy radical list
	# H2Oi - index of water
	# temp - temperature in chamber (K)
	# tempt - times that temperatures reached (s)
	# Pnow - pressure inside chamber (Pa)
	# light_stat - lights status
	# light_time - time light status attained (s)
	# daytime - time of day experiment starts (s)
	# lat - latitude of experiment (degrees)
	# lon - longitude of experiment (degrees)
	# af_path - path to actinic flux values
	# dayOfYear - number of days since 31st December experiment held on
	# photo_path - path to file with absorption cross-sections
	# 		and quantum yields
	# Jlen - number of photochemical reactions
	# con_infl_C - influx of components with constant concentration
	# (molecules/cc/s)
	# nrec_step - number of recording steps
	# dydt_vst - dictionary for holding change tendencies of specified
	#		components
	# siz_str - the size structure
	# num_sb - number of particle size bins
	# num_comp - number of components
	# seedi - index of component(s) comprising seed material
	# seed_name - name(s) of component(s) comprising seed particles
	# seedVr - volume ratio of component(s) comprising seed material
	# core_diss - dissociation constant of seed
	# Psat - pure component saturation vapour pressure 
	# 	(molecules/cc (air))
	# mfp - mean free path (m)
	# accom_coeff - accommodation coefficient
	# y_mw - molecular weight (g/mol)
	# surfT - surface tension (g/s2)
	# R_gas - ideal gas constant (kg.m2.s-2.K-1.mol-1)
	# NA - Avogadro's constant (molecules/mol)
	# y_dens - component densities (kg/m3)
	# x - particle radii (um)
	# Varr - particle volume (um3)
	# therm_sp - thermal speed (m/s)
	# act_coeff - activity coefficient
	# Cw - effective absorbing mass of wall (molecules/cc (air))
	# kw - gas-wall mass transfer coefficient (/s)
	# Cfactor - conversion factor for concentrations (ppb/molecules/cc)
	# tf - transmission factor for natural sunlight
	# light_ad - marker for whether to adapt time interval for 
	#	changing natural light intensity
	# y_arr - index for arranging concentrations into matrix that 
	# 	allows reaction rate coefficient calculation
	# y_rind - index for the concentration array  that 
	# 	allows reaction rate coefficient calculation
	# uni_y_rind - unique index of reactants
	# y_pind - index for the concentration array  that 
	# 	allows product gains to be indexed
	# uni_y_pind - unique index of products
	# reac_col - column indices for sparse matrix of reaction 
	#		losses
	# prod_col - column indices for sparse matrix of production
	#		gains
	# rstoi_flat - 1D array of reactant stoichiometries per 
	#	equation
	# pstoi_flat - 1D array of product stoichiometries per
	#	equation
	# rr_arr - index for reaction rates to allow reactant loss
	# 	calculation
	# rr_arr_p - index for reaction rates to allow product gain
	#	calculation
	# rowvals - row indices of Jacobian elements
	# colptrs - indices of  rowvals corresponding to each column of
	# 	the Jacobian
	# wall_on - marker for whether wall partitioning turned on
	# Cw - effective absorbing mass concentration of wall 
	#	(molecules/cc (air))
	# kw - mass transfer coefficient of wall partitioning (/s)
	# jac_wall_indx - index of inputs to Jacobian from wall 
	# 	partitioning
	# jac_part_indx - index of inputs to Jacobian from particle
	#	partitioning
	# Vbou - volume boundary of size bins (um3)
	# N_perbin - number concentration of particles per size bin 
	#	(#/cc (air))
	# Vol0 - initial single particle volumes per size bin (um3)
	# rad0 - initial radius at particle centres (um)
	# np_sum - number concentration of newly nucleated particles 
	#		(#/cc (air))
	# new_partr - radius of newly nucleated particles (cm)
	# nucv1, v2, v3 - nucleation parameters
	# nuc_comp - the nucleating component
	# nuc_ad - marker for whether to reduce time step to allow 
	#	for accurate capture of nucleation
	# RH - relative humidity
	# coag_on - whether coagulation to be modelled
	# inflectDp - particle diameter at which wall loss inflection occurs (m)
	# pwl_xpre - x value preceding inflection point
	# pwl_xpro - x value proceeding inflection point
	# inflectk - deposition rate at inflection (/s)
	# chamR - spherical-equivalent radius of chamber (m2)
	# McMurry_flag - marker for treament of particle deposition to walls
	# p_char - average number of charges per particle (/particle)
	# e_field - average electric field inside chamber (g.m/A.s3)
	# injectt - time of injection of components (s)
	# inj_indx - index of component(s) being injected after 
	#	experiment start
	# Ct - concentration(s) (ppb) of component(s) injected 
	#	instantaneously after experiment start
	# pmode - whether number size distributions expressed as modes or explicitly
	# pconc - concentration of injected particles (#/cc (air))
	# pconct - times of particle injection (s)
	# mean_rad - mean radius for particle number size 
	#	distribution (um)
	# lowsize - lower size bin boundary (um)
	# uppsize - upper size bin boundary (um)
	# std - standard deviation for injected particle number size 
	#	distributions
	# rbou - size bin radius bounds (um)
	# const_infl_t - times for constant influxes (s)
	# MV - molar volume (cc/mol)
	# rindx_aq - index of reactants for aqueous-phase 
	# pindx_aq - index of products for aqueous-phase
	# rstoi_aq - stoichiometry of reactants for aqueous-phase
	# pstoi_aq - stoichiometry of products for aqueous-phase
	# nreac_aq - number of reactants per aqueous-phase reaction
	# nprod_aq - number of products per aqueous-phase reaction
	# jac_stoi_aq - stoichiometry for Jacobian for aqueous-phase
	# njac_aq - number of Jacobian elements per aqueous-phase reaction 
	# jac_den_indx_aq - index of Jacobian denominators 
	# jac_indx_aq - index of  Jacobian for aqueous-phase
	# y_arr_aq - y indices for aqueous-phase
	# y_rind_aq - reactant indices for aqueous-phase 
	# uni_y_rind_aq - y indices for reactants for aqueous-phase
	# y_pind_aq - y indices for products for aqueous-phase
	# uni_y_pind_aq - y indices for products for aqueous-phase
	# reac_col_aq - columns of sparse matrix for aqueous-phase
	# prod_col_aq - columns of sparse matrix for aqueous-phase
	# rstoi_flat_aq - reactant stoichiometries for Jacobian for 
	# 	aqueous-phase
	# pstoi_flat_aq - product stoichiometries for Jacobian for
	#	aqueous-phase
	# rr_arr_aq - aqueous-phase reaction rate indices
	# rr_arr_p_aq - aqueous-phase reaction rate indices
	# eqn_num - number of reactions in gas- and aqueous-phase
	# partit_cutoff - the product of saturation vapour pressure
	#	and activity coefficient above which gas-particle
	#	partitioning assumed negligible
	# diff_vol - diffusion volumes of components according to 
	#		Fuller et al. (1969)
	# DStar_org - diffusion coefficient of components at initial temperature (cm2/s)
	# corei - index of core component
	# ser_H2O - whether to serialise the gas-particle partitioning of water
	# ------------------------------------------------------------
	
	step_no = 0 # track number of time steps
	sumt = 0.0 # track time through simulation (s)
	# counters on updates
	light_time_cnt = 0 # light time status count
	gasinj_cnt = 0 # count on injection times of components
	if pconct[0, 0] == 0.: 
		seedt_cnt = 1 # count on injection times of particles
	else:
		seedt_cnt = 0
	infx_cnt = 0 # count on constant gas-phase influx occurrences
	save_cnt = 1 # count on recording results
	# count on time since update to integration initial values/constants last called (s)
	update_count = 0.
	y0 = np.zeros((len(y))) # remember initial concentrations (molecules/cc (air))
	t0 = update_stp # remember initial integration step (s)
	# flag for changing integration time step due to changing initial values	
	ic_red = 0
	tnew = update_stp # the time to integrate over (s)
	
	# prepare recording matrices, including recording of initial
	# conditions
	[trec, yrec, dydt_vst, Cfactor_vst, Nres_dry, Nres_wet, x2, 
	seedt_cnt, rbou_rec, Cfactor, infx_cnt] = rec_prep.rec_prep(nrec_steps, 
	y, rindx, 
	rstoi, pindx, pstoi, nprod, dydt_vst, nreac, 
	num_sb, num_comp, N_perbin, core_diss, Psat, mfp,
	accom_coeff, y_mw, surfT, R_gas, temp, tempt, NA,
	y_dens*1.e-3, x, therm_sp, H2Oi, act_coeff,
	RO2_indx, sumt, Pnow, light_stat, light_time, 
	light_time_cnt, daytime, lat, lon, af_path, 
	dayOfYear, photo_path, Jlen, Cw, kw, Cfactor, tf, 
	light_ad, wall_on, Vbou, tnew, nuc_ad, nucv1, nucv2, nucv3, 
	np_sum, update_stp, update_count, injectt, gasinj_cnt, 
	inj_indx, Ct, pmode, pconc, pconct, seedt_cnt, mean_rad, corei, 
	seed_name, seedVr, lowsize, uppsize, rad0, x, std, rbou, const_infl_t, 
	infx_cnt, con_infl_C, MV, partit_cutoff, diff_vol, DStar_org, seedi)

	print('Starting loop through update steps')	
	while (tot_time-sumt)>(tot_time/1.e10):
		
		y0[:] = y[:] # remember initial concentrations (molecules/cc (air))
		
		# update chamber variables
		[temp_now, Pnow, lightm, light_time_cnt, tnew, ic_red, update_stp, 
			update_count, Cinfl_now, seedt_cnt, Cfactor, infx_cnt, 
			gasinj_cnt, DStar_org] = cham_up.cham_up(sumt, temp, tempt, 
			Pnow, light_stat, light_time, light_time_cnt, light_ad, 
			tnew, nuc_ad, nucv1, nucv2, nucv3, np_sum, 
			update_stp, update_count, lat, lon, dayOfYear, photo_path, 
			af_path, injectt, gasinj_cnt, inj_indx, Ct, pmode, pconc, pconct, 
			seedt_cnt, num_comp, y, N_perbin, mean_rad, corei, seedVr, seed_name, 
			lowsize, uppsize, num_sb, MV, rad0, x, std, y_dens, H2Oi, rbou, 
			const_infl_t, infx_cnt, con_infl_C, wall_on, Cfactor, seedi, diff_vol, DStar_org)
		
		# ensure end of time interval does not surpass recording time
		if ((sumt+tnew) > save_stp*save_cnt):
			tnew = (save_stp*save_cnt)-sumt
			ic_red = 1

		# ensure update step interval not surpassed
		if (update_count+tnew > update_stp):
			tnew = (update_stp-update_count)
			ic_red = 1
			
		# ensure simulation end time not surpassed
		if (sumt+tnew > tot_time):
			tnew = (tot_time-sumt)
			ic_red = 1
		
		if ((num_sb-wall_on) > 0 and (sum(N_perbin) > 0)): # if particles present
			# update partitioning variables
			[kimt, kelv_fac] = partit_var.kimt_calc(y, mfp, num_sb, num_comp, accom_coeff, 
			y_mw, surfT, R_gas, temp_now, NA, y_dens, N_perbin, 
			x.reshape(1, -1)*1.0e-6, Psat, therm_sp, H2Oi, act_coeff, wall_on, 1, partit_cutoff, 
			Pnow, DStar_org)
						
		else: # fillers
			kimt = kelv_fac = 0.
		
		# reaction rate coefficient
		rrc = rrc_calc.rrc_calc(RO2_indx, 
			y[H2Oi], temp_now, lightm, y, daytime+sumt, 
			lat, lon, af_path, dayOfYear, Pnow, 
			photo_path, Jlen, tf)
		
		# update Jacobian inputs based on particle-phase fractions of components
		[rowvalsn, colptrsn, jac_part_indxn, jac_mod_len, jac_part_hmf_indx, rw_indx, jac_wall_indxn, jac_part_H2O_indx] = jac_up.jac_up(y[num_comp:num_comp*((num_sb-wall_on+1))], rowvals, colptrs, (num_sb-wall_on), num_comp, jac_part_indx, H2Oi, y[H2Oi], jac_wall_indx, ser_H2O)
		
		if (ser_H2O == 1 and (num_sb-wall_on) > 0 and (sum(N_perbin) > 0)): # if water gas-particle partitioning serialised
		
			# call on ode solver for water
			[y, res_t] = ode_solv_wat.ode_solv(y, tnew, rindx, pindx, rstoi, pstoi,
			nreac, nprod, rrc, jac_stoi, njac, jac_den_indx, jac_indx,
			Cinfl_now, y_arr, y_rind, uni_y_rind, y_pind, uni_y_pind, 
			reac_col, prod_col, rstoi_flat, 
			pstoi_flat, rr_arr, rr_arr_p, rowvalsn, colptrsn, num_comp, 
			num_sb, wall_on, Psat, Cw, act_coeff, kw, jac_wall_indxn,
			seedi, core_diss, kelv_fac, kimt, (num_sb-wall_on), 
			jac_part_indxn,
			rindx_aq, pindx_aq, rstoi_aq, pstoi_aq,
			nreac_aq, nprod_aq, jac_stoi_aq, njac_aq, jac_den_indx_aq, jac_indx_aq, 
			y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, uni_y_pind_aq, 
			reac_col_aq, prod_col_aq, rstoi_flat_aq, 
			pstoi_flat_aq, rr_arr_aq, rr_arr_p_aq, eqn_num, jac_mod_len, 
			jac_part_hmf_indx, rw_indx, N_perbin, jac_part_H2O_indx, H2Oi)
			
			# zero partitioning of water to particles for integration without water gas-particle partitioning
			kimt[:, H2Oi] = 0.

		# model component concentration changes to get new concentrations
		# (molecules/cc (air))
		[y, res_t] = ode_solv.ode_solv(y, tnew, rindx, pindx, rstoi, pstoi,
			nreac, nprod, rrc, jac_stoi, njac, jac_den_indx, jac_indx,
			Cinfl_now, y_arr, y_rind, uni_y_rind, y_pind, uni_y_pind, 
			reac_col, prod_col, rstoi_flat, 
			pstoi_flat, rr_arr, rr_arr_p, rowvalsn, colptrsn, num_comp, 
			num_sb, wall_on, Psat, Cw, act_coeff, kw, jac_wall_indxn,
			seedi, core_diss, kelv_fac, kimt, (num_sb-wall_on), 
			jac_part_indxn,
			rindx_aq, pindx_aq, rstoi_aq, pstoi_aq,
			nreac_aq, nprod_aq, jac_stoi_aq, njac_aq, jac_den_indx_aq, jac_indx_aq, 
			y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, uni_y_pind_aq, 
			reac_col_aq, prod_col_aq, rstoi_flat_aq, 
			pstoi_flat_aq, rr_arr_aq, rr_arr_p_aq, eqn_num, jac_mod_len, 
			jac_part_hmf_indx, rw_indx, N_perbin, jac_part_H2O_indx, H2Oi)

		step_no += 1 # track number of steps
		sumt += tnew # total time through simulation (s)
		
		if (num_sb-wall_on > 0): # if particle size bins present
			# update particle sizes
			if ((num_sb-wall_on) > 1) and (any(N_perbin > 1.e-10)): # if particles present
				
				if (siz_str == 0): # moving centre
					(N_perbin, Varr, y, x, redt, t, bc_red) = mov_cen.mov_cen_main(N_perbin, 
					Vbou, num_sb, num_comp, y_mw, x, Vol0, tnew, 
					update_stp, y0, MV, Psat[0, :], ic_red, y, res_t, wall_on)
					
				if (siz_str == 1): # full-moving
					(Varr, x, y[num_comp:(num_comp*(num_sb-wall_on+1))], 
					N_perbin, Vbou, rbou) = fullmov.fullmov((num_sb-wall_on), N_perbin,
 					num_comp, y[num_comp:(num_comp)*(num_sb-wall_on+1)], MV*1.e12, 
					Vol0, Vbou, rbou)
			
			update_count += tnew # time since operator-split processes last called (s)

			# if time met to implement operator-split processes
			if (update_count>=update_stp*9.999999e-1):
				if (any(N_perbin > 1.e-10)):
				
					# particle-phase concentration(s) (molecules/cc (air))
					Cp = np.transpose(y[num_comp:(num_comp)*(num_sb-wall_on+1)].reshape(num_sb-wall_on, num_comp))
					
					# coagulation
					[N_perbin, y[num_comp:(num_comp)*(num_sb-wall_on+1)], x, Gi, eta_ai, 
						Varr, Vbou, rbou] = coag.coag(RH, temp_now, x*1.e-6, 
						(Varr*1.0e-18).reshape(1, -1), 
						y_mw.reshape(-1, 1), x*1.0e-6, 
						Cp, (N_perbin).reshape(1, -1), update_count, 
						(Vbou*1.0e-18).reshape(1, -1), rbou,
						num_comp, 0, (np.squeeze(y_dens*1.0e-3)), Vol0, rad0, Pnow, 0,
						Cp, (N_perbin).reshape(1, -1), (Varr*1.0e-18).reshape(1, -1),
						coag_on, siz_str, wall_on)
					
					if ((McMurry_flag > -1) and (wall_on == 1)): #if particle loss to walls turned on
						# particle loss to walls
						[N_perbin, 
						y[num_comp:(num_comp)*(num_sb-wall_on+1)]] = wallloss.wallloss(
							N_perbin.reshape(-1, 1), 
							y[num_comp:(num_comp)*(num_sb-wall_on+1)], Gi, eta_ai,
 							x*2.0e-6, y_mw, 
							Varr*1.0e-18, num_sb, num_comp, temp_now, update_count, 
							inflectDp, pwl_xpre, pwl_xpro, inflectk, chamR, McMurry_flag, 
							0, p_char, e_field, (num_sb-wall_on))
			
				if (nucv1 > 0.): # nucleation
					
					[N_perbin, y, x, Varr, np_sum, rbou, Vbou] = nuc.nuc(sumt, np_sum, 
						N_perbin, y, y_mw.reshape(-1, 1), 
						np.squeeze(y_dens*1.0e-3),  
						num_comp, Varr, x, new_partr, MV, nucv1, nucv2, 
						nucv3, nuc_comp[0], siz_str, rbou, Vbou, (num_sb-wall_on))
				
				# reset count to that since original operator-split processes interval met (s)
				update_count = sumt%t0

		
		print('time through simulation (s): ', sumt)
		
		# record output
		if ((sumt-(save_stp*save_cnt) > -1.e-10) or (sumt >= (tot_time-tot_time/1.e10))):
			
			[trec, yrec, dydt_vst, Cfactor_vst, save_cnt, 
				Nres_dry, Nres_wet, x2, rbou_rec] = rec.rec(save_cnt, trec, yrec, 
				dydt_vst, Cfactor_vst, y, sumt, rindx, rstoi, rrc, pindx, pstoi, 
				nprod, nreac, num_sb, num_comp, N_perbin, core_diss, 
				Psat, kelv_fac, kimt, kw, Cw, act_coeff, Cfactor, Nres_dry, 
				Nres_wet, x2, x, MV, H2Oi, Vbou, rbou, wall_on, rbou_rec, seedi)		
		
		# if time step was temporarily reduced, then return
		if (ic_red == 1):
			update_stp = t0
			tnew = update_stp
			ic_red = 0
	
	return(trec, yrec, dydt_vst, Cfactor_vst, Nres_dry, Nres_wet, x2, rbou_rec)



'''code to open saved PyCHAM files and return useful variables'''
# called by PyCHAM plotting codes to obtain the required model outputs for evaluation

import numpy as np

def retr_out(output_by_sim):
	
	# inputs: -------------------------------
	# output_by_sim - name of folder requested by the calling code to be looked at
	# ---------------------------------------
	
	# name of file where experiment constants saved
	fname = str(output_by_sim + '/model_and_component_constants')
	
	const_in = open(fname)

	const = {} # prepare to create dictionary
	for line in const_in.readlines():
		# convert to python list
		dlist = []
		for i in line.split(',')[1::]:
			
			if str(line.split(',')[0]) == 'number_of_size_bins':
				dlist.append(int(i))
			if str(line.split(',')[0]) == 'number_of_components' or str(line.split(',')[0]) == 'wall_on_flag_0forNO_1forYES':
				dlist.append(int(i))
			if str(line.split(',')[0]) == 'molecular_weights_g/mol_corresponding_to_component_names' or  str(line.split(',')[0]) == 'molecular_volumes_cm3/mol' or  str(line.split(',')[0]) == 'molar_volumes_cm3/mol':
				i = i.strip('\n')
				i = i.strip('[')
				i = i.strip(']')
				i = i.strip(' ')
				dlist.append(float(i))
			if str(line.split(',')[0]) == 'component_names' or str(line.split(',')[0]) == 'space_mode':
				i = i.strip('\n')
				i = i.strip('[')
				i = i.strip(']')
				i = i.strip(' ')
				i = i.strip('\'')
				dlist.append(str(i))
			if str(line.split(',')[0]) == 'factor_for_multiplying_ppb_to_get_molec/cm3_with_time':
				i = i.strip('\n')
				i = i.strip(' ')				
				i = i.strip('[[')
				i = i.strip('[')
				i = i.strip(']')
				dlist.append(float(i))
			if str(line.split(',')[0]) == 'simulation_computer_time(s)':
				i = i.strip('\n')
				i = i.strip('[')
				i = i.strip(']')
				i = i.strip(' ')
				dlist.append(float(i))
			
		const[str(line.split(',')[0])] = dlist
	const_in.close()
	num_sb = int((const['number_of_size_bins'])[0]) # number of size bins
	num_speci = int((const['number_of_components'])[0]) # number of species
	# conversion factor to change gas-phase concentrations from molecules/cc 
	# (air) into ppb 
	Cfactor = const['factor_for_multiplying_ppb_to_get_molec/cm3_with_time']
	PyCHAM_names = const['component_names']
	y_MW = const['molecular_weights_g/mol_corresponding_to_component_names']
	spec_namelist = const['component_names']
	wall_on = const['wall_on_flag_0forNO_1forYES'][0]
	space_mode = const['space_mode'][0]
	
	try:
		MV = const["molecular_volumes_cm3/mol"]
	except:
		MV = const["molar_volumes_cm3/mol"]

	try:
		speed = (const["simulation_computer_time(s)"])[0]
	except:
		speed = 10.0
		
	# withdraw times (s)
	fname = str(output_by_sim+'/time')
	t_array = np.loadtxt(fname, delimiter=',', skiprows=1)
	timehr = t_array/3600.0 # convert from s to hr
	
	# withdraw concentrations
	fname = str(output_by_sim+'/concentrations_all_components_all_times_gas_particle_wall')
	y = np.loadtxt(fname, delimiter=',', skiprows=1)
	
	try:
		# withdraw number-size distributions (# particles/cc (air))
		fname = str(output_by_sim+'/particle_number_concentration_dry')
		N = np.loadtxt(fname, delimiter=',', skiprows=1)
	except:
		N = []
	
	try:
		# withdraw number-size distributions (# particles/cc (air))
		fname = str(output_by_sim+'/particle_number_concentration_wet')
		Nwet = np.loadtxt(fname, delimiter=',', skiprows=1)
	except:
		Nwet = []
	
	try:
		# particle sizes (um)
		fname = str(output_by_sim+'/size_bin_radius')
		x = np.loadtxt(fname, delimiter=',', skiprows=1) # skiprows=1 omits header
	except:
		x = []

	try:
		# particle size bin bounds (radii) (um3)
		fname = str(output_by_sim+'/size_bin_bounds')
		rbou_rec = np.loadtxt(fname, delimiter=',', skiprows=1) # skiprows=1 omits header
	except:
		rbou_rec = []
	
	return(num_sb, num_speci, Cfactor, y, N, rbou_rec, x, timehr, PyCHAM_names, y_MW, 
		Nwet, spec_namelist, MV, speed, wall_on, space_mode)



''' module to estimate variables related to the fluidity of particles in air'''

# eq. nos. refer to Jacobson (2005)

import numpy as np 
import scipy.constants as si
import pdb

def reg_determ(RH, T, sbr, Pa):

	# ------------------------------------------------
	# inputs:
	# RH - relative humidity at this time (fraction) 
	# T - temperature (K)
	# sbr - size bin radii (m)
	# Pa - pressure inside chamber (Pa)
	
	# ------------------------------------------------
	# outputs:
	# Kni - Knudsen number of particles
	# rho_a - mass density of moist air
	# eta_a - dynamic viscosity (4.54) (g/m.s)
	# kin_visc - kinematic viscosity of air (4.55) (m2/s)	
	# ------------------------------------------------

	# saturation vapour pressure of water vapour (Pa (kg/m.s2) 
	# (2.61))
	Pvs = (6.112e2*np.exp(6816*(1.0/273.15-1.0/T)+
		(5.1309*np.log(273.15/T))))
	
	# vapour pressure (partial pressure) of water vapour 
	# (Pa (kg/m.s2) (2.66)
	Pv = (RH*Pa*Pvs)/(RH*Pvs+Pa-Pvs)
	# pressure of dry air (Pa) (2.22)
	Pd = Pa-Pv
	
	# ratio of molecular weights of water and dry air (p. 30 for Mw air
	# & CRC for Mw water vapour) (2.31)
	epsilon = (18.015/28.966)
	# mass mixing ratio of water vapour (2.31)
	omega = epsilon*(Pv/Pd)
	# dry-air gas constant (m2/s2.K)  
	# (2.24 (R* value from p. 29 (converted to units 
	# m2/s.K}))
	R_dash = (8314.51/28.966)
	# gas constant for moist air (2.37) (m3Pa/g.K)	
	Rm = R_dash*((1.0+omega/epsilon)/(1.0+omega))	
	# mass density of moist air (g/m3) (2.36 Jacobson 2005)
	# (multiply by 1.0e3) to convert from kg/m3 to g/m3
	rho_a = Pa/(Rm*T)*1.0e3
	
	
	# dynamic viscosity of air (g/m.s) (4.54) 
	# (p. 30 for Mw air and p. 102 for av. diameter of air molecule (m))
	eta_a = (5.0/(16.0*si.N_A*(3.673e-10**2.0)))*(((28.966*8314.51*T)
		/(np.pi))**0.5)
	# kinematic viscosity of air (m2/s) (4.55)
	kin_visc = eta_a/rho_a

	# thermal speed of air molecule (m/s) (2.3) (15.25) 
	# (4.8096e-26 is average mass of one air molecule (p. 18))
	nu_a = ((8.0*si.k*T)/(np.pi*4.8096e-26))**(0.5)	
	# mean free path of an air molecule (m) (15.24)
	lamb = (2.0*kin_visc)/nu_a

	# Knudsen number for particles in air (15.23) - determines flow regime
	Kni = lamb/sbr	
		
	return Kni, eta_a, rho_a, kin_visc




'''module to set up particle phase part of box model, calls on init_water_partit to initiate water partitioning with seed particles and wall'''
# gets called when there is an injection of seed particle during an experiment

import numpy as np
from init_water_partit import init_water_partit
import scipy.constants as si
from scipy import stats # import the scipy.stats module

def pp_dursim(y, N_perbin, mean_rad, pmode, pconc, seedi, seedVr, lowersize, uppersize, num_comp, 
				num_sb, MV, rad0, radn, std, y_dens, H2Oi, rbou):
	
			
	# inputs -----------------------------------
	# y - concentrations of components in particle phase (molecules/cc (air))
	# N_perbin - number concentration of particles (#/cc (air))
	# mean_rad - mean radius of seed particles at this time (um)
	# pmode - whether particle number size distribution stated by mode or explicitly
	# pconc - number concentration of seed particles (#/cc (air))
	# seedi - index of seed material
	# seedVr - volume ratio of component(s) comprising seed particles
	# lowersize - smallest radius bound (um)
	# uppersize - greatest radius bound (um)
	# num_comp - number of components
	# num_sb - number of size bins (excluding wall)
	# MV - molar volume (cc/mol)
	# rad0 - original radius at size bin centres (um)
	# radn - current radius at size bin centres (um)
	# std - standard deviation for lognormal size distribution calculation (dimensionless)
	# y_dens  - density of components (kg/m3)
	# H2Oi - index of water
	# rbou - radius bounds per size bin (um)
	# ------------------------------------------
	
	
	# if mean radius not stated explicitly calculate from size ranges (um)
	if (mean_rad == -1.e6 and num_sb > 0):
		if (lowersize > 0.):
			mean_rad = 10**((np.log10(lowersize)+np.log10(uppersize))/2.)
		else:
			mean_rad = 10**((np.log10(uppersize))/2.)
	
	R_gas = si.R # ideal gas constant (kg.m2.s-2.K-1.mol-1)
	NA = si.Avogadro # Avogadro's number (molecules/mol)
	
	if num_sb == 1:
		N_perbin += np.array((pconc)) # (# particles/cc (air))
		pconc_new = pconc

	# number concentration stated per size bin in multi size bin simulation
	if (pmode == 1): 
		N_perbin += np.array((pconc)) # (# particles/cc (air))
		pconc_new = pconc

	# total number concentration per mode stated in multi size bin simulation
	if (pmode == 0):
		
		for i in range(len(pconc)): # loop through modes
			# set scale and standard deviation input for lognormal probability distribution 
			# function, following guidance here: 
			# http://all-geo.org/volcan01010/2013/09/how-to-use-lognormal-distributions-in-python/
			scale = np.exp(np.log(mean_rad[i]))
			std = np.log(std[i])
			loc = 0.0 # no shift
		
			# number fraction-size distribution - enforce high resolution to ensure size
			# distribution of seed particles fully captured
			if (lowersize > 0.): # if lowermost size bin bound useful
				hires = 10**(np.linspace(np.log10(lowersize), np.log10(uppersize), int(num_sb*1.e2)))
			else: # enforce lowermost size bin radius bound of 1 nm (1e-3 um)
				hires = 10**(np.linspace(np.log10(1.e-3), np.log10(uppersize), int(num_sb*1.e2)))
			pdf_output = stats.lognorm.pdf(hires, std, loc, scale)
			pdf_out = np.interp(radn, hires, pdf_output)	
			# number concentration of seed in all size bins (# particle/cc (air))
			pconc_new = (pdf_out/sum(pdf_out))*pconc[i]
			# number concentration realism
			pconc_new[pconc_new<1] = 0.
			N_perbin[:, 0] += pconc_new # (# particles/cc (air))
	
	# molecular concentration of seed required to comprise these additional seed particle
	# (molecules/cc (air)):
	
	# volume concentration of seed particles (um3/cc (air))
	Vperbin = ((pconc_new*(4.0/3.0)*np.pi*(radn)**3.0))

	if (sum(pconc_new) > 0.): # account for concentration of component(s) comprising seed
		for ci in range(len(seedi)): # loop through indices of seed components 
			# concentration in all size bins (molecules/cc (air)):
			y[seedi[ci]:num_comp*num_sb:num_comp] += (NA/(MV[seedi[ci]]*1.e12))*(Vperbin*(seedVr[ci]/sum(seedVr)))

	# loop through size bins to estimate new total volume concentrations (um3/cc (air))
	Vtot = np.zeros((num_sb))
	Varr = np.zeros((num_sb)) # volume concentration of single particles (um3/cc (air))
	mass_conc = 0. # mass concentration of particles (g/cm3)
	
	for i in range(num_sb): # size bin loop
		Vtot[i] = np.sum(y[num_comp*i:num_comp*(i+1)]/NA*(MV*1.e12))
		# new volume concentration of single particles (um3/cc (air))
		if (N_perbin[i] > 0.):
			Varr[i] = Vtot[i]/N_perbin[i]
		else:
			Varr[i] = (4./3.*np.pi)*rad0[i]**3.
		# multiply y_dens by 1e-3 to get ug/um3 (particle) from kg/m3, 
		# then multiplying ug/um3 (particle) by um3/cc (air) gives ug/cc (air)
		mass_conc += np.sum((y_dens[seedi, 0]*1.0e-9)*Vperbin[i])
	
	# scale mass_conc by 1e6 to convert from ug/cm3 (air) to ug/m3 (air) 
	print(str('Total dry (no water) mass concentration of newly injected seed particles: ' + str(mass_conc*1.e6) + ' ug/m3 (air)'))

	# new radius of single particles (um)
	radn = ((3.0/(4.0*np.pi))*Varr)**(1.0/3.0)
	
	return(y, N_perbin, radn, Varr)



'''module to prepare PyCHAM for partitioning variable calculation (particle and wall)'''

# module responsible for preparing inputs to the calculation of the gas-particle 
# partitioning coefficient in kimt_calc

import numpy as np
import scipy.constants as si
import diff_vol_est

def prep(y_mw, TEMP, num_speci, testf, Cw, act_comp, act_user, acc_comp, 
			accom_coeff_user, spec_namelist, num_sb, num_asb, Pnow, 
			Pybel_object, name_SMILE):
	
	# ------------------------------------------------------------------
	# inputs:
	# y_mw - molecular weight of components (g/mol) (num_speci,1)
	# TEMP - temperature of chamber at start of experiment (K)
	# num_speci - number of components
	# testf - flag for whether in normal mode (0) or testing mode (1)
	# Cw - effective absorbing mass of wall (g/m3 (air))
	# act_comp - names of components (corresponding to chemical scheme name) with 
	# 			activity coefficient stated in act_user
	# act_user - user-specified activity coefficients of components with names given in
	#			act_comp
	# accom_comp - names of components with accommodation coefficient set by the user
	# accom_coeff_user - accommodation coefficient set by the user
	# spec_namelist - names of components as stated in the chemical scheme
	# num_sb - number of size bins (excluding wall)
	# num_asb - number of actual size bins excluding wall
	# Pnow - air pressure inside chamber (Pa)
	# Pybel_object - Pybel objects for components
	# name_SMILE - SMILE strings of components
	# -----------------------------------------------------------------
	
	if testf == 1: # if in testing mode (for test_front.py)
		return(0,0,0,0,0,0,0) # return dummies
	
	surfT = 72. # assume surface tension of water (g/s2==mN/m==dyn/cm) for all particles
	
	# dynamic viscosity of air (kg/m.s), eq. 4.54 of Jacobson 2005
	dyn_visc = 1.8325e-5*((416.16/(TEMP+120.))*(TEMP/296.16)**1.5)
	
	ma = 28.966e-3 # molecular weight of air (kg/mol) (eq. 16.17 Jacobson 2005)
	
	# air density (kg/m3 (air)), ideal gas law
	rho_a =  (Pnow*ma)/((si.R)*TEMP)
	
	# mean thermal speed of each component (m/s) (eq. 2.3 Jacobson 2005)
	# note that we need the weight of one molecule, which is why y_mw is divided by
	# Avogadro's constant, and we need it in kg, which is why we multiply by 1e-3
	therm_sp = ((8.*si.k*TEMP)/(np.pi*(y_mw/si.N_A)*1.e-3))**0.5
	
	# mean free path (m) for each component (15.24 of Jacobson 2005)
	mfp = (2.*dyn_visc/(rho_a*therm_sp)).reshape(-1, 1)

	nv = (Pnow/(si.R*TEMP))*si.N_A # concentration of molecules (# molecules/m3)
	
	# get diffusion volumes
	diff_vol = diff_vol_est.diff_vol_est(Pybel_object)
	
	# append water and core (water from Table 4.1 of the Taylor (1993) textbook 
	# Multicomponent Mass Transfer, ISBN: 0-471-57417-1)
	diff_vol = (np.append(diff_vol, np.array((13.1, 1.)))).reshape(-1, 1)
	
	# diffusion coefficient (m2/s) of components in gas phase (air), eq 4.1.4 of
	# the Taylor (1993) textbook 
	# Multicomponent Mass Transfer, ISBN: 0-471-57417-1, note diffusion volume for air 
	# (19.7) taken from Table 4.1 of Taylor (1993) and mw of air converted to g/mol from 
	# kg/mol.  This is a replication of the original method from Fuller et al. (1969): 
	# doi.org/10.1021/j100845a020
	Dstar_org = 1.013e-2*TEMP**1.75*(((y_mw+ma*1.e3)/(y_mw*ma*1.e3))**0.5)/(Pnow*(diff_vol**(1./3.)+19.7**(1./3.))**2.)
	
	# convert to cm2/s
	Dstar_org = Dstar_org*1.e4
	
	# accommodation coefficient of components in each size bin
	accom_coeff = np.ones((num_speci, num_sb))*1.e0
	
	# list containing accommodation coefficients that are functions
	accom_coeff_func = []
	
	ac_indx = []
	for i in range(len(acc_comp)): # user-defined accommodation coefficients
		# get index of component stated
		ac_indx.append(spec_namelist.index(acc_comp[i].strip()))

	# check for any accommodation coefficients set by user
	if len(ac_indx)>0:
		for i in range(len(ac_indx)):

			# if it is a constant (not a function, which would be a string)
			if isinstance(accom_coeff_user[i], str) == False:
				accom_coeff[ac_indx] = accom_coeff_user[i]
			# if it is a function, it will be a string and needs making available to the 
			# kimt_calc module
			if isinstance(accom_coeff_user[i], str) == True:
				accom_coeff_func.append(str('accom_coeff[' + str(ac_indx[i]) + ', :]' + ' = ' + accom_coeff_user[i]))
	
	# generate module that contains any accommodation coefficient functions, note, do 
	# this even if no functions supplied so that the accomm_coeff_calc is updated and
	# accurate for this simulation
	f = open('PyCHAM/accom_coeff_calc.py', mode='w')
	f.write('\'\'\'module for calculating accommodation coefficients, automatically generated by kimt_prep\'\'\'\n')
	f.write('\n')
	f.write('# code that expresses and performs the functions for accommodation \n')
	f.write('# coefficients that are given by the user in the model variables file \n')
	f.write('# and that are intended to be calculated real-time via the kimt_calc \n')
	f.write('# module \n')
	f.write('\n')

	# following part is the function (there should be an indent at the start of each line
	# following the def line - suggest using 1 tab)
	f.write('def accom_coeff_func(accom_coeff, radius):\n')
	f.write('\n')
	f.write('	# ------------------------------------------------------------------ \n')
	f.write('	# inputs:\n')
	f.write('	# accom_coeff - array containing accommdotion coefficients for all \n')
	f.write('	# components\n')
	f.write('	# radius - radii of size bins (m)\n')
	f.write('	# ------------------------------------------------------------------ \n')
	f.write('\n')
	f.write('	# calculate accommodation coefficients \n')
	# code to calculate accommodation coefficients as given by user 
	for line in accom_coeff_func:
		f.write('	%s \n' %line)
	f.write('\n')
	f.write('	return(accom_coeff)\n')
	f.close()
	
	# activity coefficient of components - affects the particle- and wall-phase
	act_coeff = np.ones((1, num_speci))
	for i in range(len(act_comp)): # user-defined activity coefficients
		# get index of component stated
		ac_indx = spec_namelist.index(act_comp[i].strip())
		act_coeff[0, ac_indx] = act_user[i].strip()
	
	# in preparation for use in ode solver, repeat activity coefficients over
	# size bins
	if num_asb>0:
		act_coeff = np.repeat(act_coeff, num_asb, axis=0)

	# convert Cw (effective absorbing mass of wall) from g/m3 (air) to 
	# molecules/cc (air), assuming a molecular weight of 200g/mol (*1.0e-6 to convert from
	# /m3 (air) to /cm3 (air))
	Cw = ((Cw*1.e-6)/200.)*si.N_A
	
	R_gas = si.R # ideal gas constant (kg.m2.s-2.K-1.mol-1)
	NA = si.Avogadro # Avogadro's constant (molecules/mol)

	return(mfp, accom_coeff, therm_sp, surfT, Cw, act_coeff, R_gas, NA, diff_vol, Dstar_org)



'''checking that user inputs are valid'''

import os
import sys
import numpy as np

def ui_check(sav_nam, sch_name, wall_on, caller, siz_str, num_sb, pmode, pconc, pconct, lowsize, std, mean_rad, new_partr, chamSA, chem_sch_mark, af_path, int_tol, update_stp, tot_time, RH, uman_up, light_stat, light_time, injectt, Ct, dens_comp, dens, seed_name, seedVr, seed_diss, partit_cutoff):

	# inputs: ------------------------------------------------------------
	# sav_nam - name of folder to save results to
	# sch_name - name of chemical scheme file
	# wall_on - marker for whether wall on or off
	# caller - marker for the calling module
	# siz_str - the size structure
	# num_sb - number of particle size bins
	# pmode - whether particle number concentrations expressed by mode or explicitly
	# pconc - number concentration of particles
	# pconct - times of particle injection (s)
	# lowsize - lower bound of particle sizes (um)
	# std - standard deviation for particle sizes
	# mean_rad - mean radius of particle size distributions (um)
	# new_partr - radius of newly nucleated particles (cm)
	# chamSA - chamber surface area (m2)
	# chem_sch_mark - markers in chemical scheme
	# af_path - actinic flux path
	# int_tol - integration tolerances
	# update_stp - time step for updating ODE initial conditions (s)
	# tot_time - total experiment time (s)
	# RH - relative humidity (0-1)
	# uman_up - marker for whether to update UManSysProp folder
	# light_stat - marker for whether lights on or off
	# light_time - time that lights attain status
	# injectt - times of instantaneous injections of gas-phase components (s)
	# Ct - concentrations of components instantaneously injected (ppb)
	# dens_comp - chemical scheme names of components with density 
	# manually assigned
	# dens - manually assigned densities (g/cc)
	# seed_name - name of component(s) comprising seed particles
	# seedVr - volume ratio of component(s) comprising seed particles
	# seed_diss - dissociation constant for seed particle component(s)
	# partit_cutoff - product of vapour pressure and activity coefficient
	# 	above which gas-particle partitioning assumed zero 
	# --------------------------------------------------------------------

	print('Checking user inputs')

	dir_path = os.getcwd() # current working directory
	output_root = 'PyCHAM/output'
	filename = os.path.basename(sch_name)
	filename = os.path.splitext(filename)[0]
	# one folder for one simulation
	output_by_sim = os.path.join(dir_path, output_root, filename, sav_nam)
	
	# constrain wall_on marker
	if (wall_on>=1):
		wall_on = 1
	else:
		wall_on = 0 

	if os.path.isdir(output_by_sim) == True and caller == 0:
		print('Error: results file name (' +output_by_sim+ ') already exists, please use an alternative')
		sys.exit()

	# ensure size structure marker is sensible
	if siz_str<0 or siz_str>1:
		siz_str = 0
	
	# consistency between number of particle size bins and particle number concentration
	if num_sb == 0 and (sum(pconc != 0) > 0):
		pconc[:] = 0.0
		print('Note that zero particle size bins registered (number_size_bins in model variables input file), however total particle number concentration (pconc in model variables input file) contains a non-zero value, therefore assuming no particle size bins')
	
	# if lower bound of particle sizes set to 0, this will induce error when taking log10
	# in pp_intro, so change to very small value (um)
	if lowsize == 0.:
		lowsize = 1.e-3

	if (num_sb > 0): # if size bins present

		if (pmode == 0): # number concentrations expressed as mode
			# if mean_rad inconsistent with pconc
			if (mean_rad.shape != pconc.shape):
				mean_rad = np.ones((pconc.shape))*-1.e6 # default flag
			if (std.shape != pconc.shape):
				std = np.ones((pconc.shape))*1.2 # default value  

	# if radius of newly nucleated particles empty, set to default value of 1 nm (cm)
	if new_partr == 0:
		new_partr = 2.e-7

	# convert chamber surface area (m2) to spherical equivalent radius (m)
	# (below eq. 2 in Charan (2018))
	chamR = (chamSA/(4.0*np.pi))**0.5

	if (len(chem_sch_mark) == 0): # if empty default to MCM kpp format
		chem_sch_mark = ['{', 'RO2', '+', 'C(ind_', ')', '', '&', '', '', ':', '}', ';']

	if not af_path: # actinic flux path
		af_path = str('no')

	if not int_tol: # integration tolerances
		int_tol = [1.e-3, 1.e-4]

	if not update_stp: # update step (s)
		update_stp = 6.e1

	if not tot_time: # total experiment time (s)
		tot_time = 3.6e3*4.

	if RH>1.0:
		print('Note: RH set above 1.0; simulation will be attempted, but please note that RH is interpreted as fraction, not a percentage, where 1 represents saturation of water vapour')

	# for UManSysProp if no update requested, check that there 
	# is an existing UManSysProp folder
	if uman_up == 0: # check for existing umansysprop folder
		if not os.path.isdir(os.getcwd() + '/umansysprop'):
			uman_up = 1
	
	if uman_up == 1: # test whether UManSysProp can be updated
		import urllib.request # module for checking internet connection
		# function for testing internet connection
		def connect(host='https://github.com/loftytopping/UManSysProp_public.git'):
			try:
				urllib.request.urlopen(host) #Python 3.x
				return True
			except:
				return False
		# test internet connection
		if connect():
			print('Internet connection confirmed and either user has requested cloning of UManSysProp via the model variables input file or no existing UManSysProp folder was found') 
		else:
			print('Error: either user has requested cloning of UManSysProp via the model variables input file or no existing UManSysProp folder has been found, but connection to the page failed, possibly due to no internet connection (UManSysProp repository site: https://github.com/loftytopping/UManSysProp_public.git)')
			sys.exit()

	# ensure that if multiple instantaneous injections of particles, that corresponding variables
	# have the correct shape, specifically that they cover the same number of times
	if (pconc.shape[1] != pconct.shape[1] or pconc.shape[1] != std.shape[1] or pconc.shape[1] != mean_rad.shape[1] or pconct.shape[1] != std.shape[1] or pconct.shape[1] != std.shape[1] or std.shape[1] != mean_rad.shape[1]):
		print(str('Error: inconsistent number of times for instantaneous injection of particles represented by model variable inputs (number of times represented in brackets) for: pconc ('+str(pconc.shape[1])+'), pconct ('+str(pconct.shape[1])+'), mean_rad ('+str(mean_rad.shape[1])+') and/or std ('+str(std.shape[1])+').  Please see README for guidance.'))
		sys.exit()
	
	if (len(light_stat) != len(light_time)):
		print('Error: length of input model variables light_status and light_time have different lengths and they should be the same, please see README for guidance.')
		sys.exit()

	
	if (light_time[0] != 0): # if no status provided at simulation start default to lights off
		light_stat = np.concatenate((np.zeros((1)).astype(int), light_stat))
		light_time = np.concatenate((np.zeros((1)), light_time))

	# check consistency between number of times components instantaneously injected and
	# the number of concentrations given
	if (len(injectt) != Ct.shape[1]):
		print('Error: number of times given for the instantaneous injection of gas-phase components (injectt in the model variables input file) is inconsistent with the number of concentrations of instantaneous injection provided (Ct in the model variables input file), please see the README for guidance')
		sys.exit()

	# check on consistency of manually assigned densities
	if (len(dens_comp) != len(dens)):
		print('Error: the number of chemical scheme names provided in the dens_Comp model variables input does not match the number of densities provided in the dens model variables input, please see README for details')
		sys.exit()

	# check on consistency of names of seed component(s) and their volume ratio
	if (len(seed_name) != len(seedVr)):
		if (len(seed_name) > 1) and (len(seedVr) == 1):
			seedVr = np.ones((len(seed_name)))
		else:
			print('Error: the number of seed particle component names (seed_name in model variables input file) and the number of seed particle component volume ratios (seedVr in model variables input file) are inconsistent, please see README for guidance')
			sys.exit()

	# check on consistency of names of seed component(s) and their dissociation constant
	if (len(seed_name) != len(seed_diss)):
		if (len(seed_name) > 1) and (len(seed_diss) == 1):
			seed_diss = np.ones((len(seed_name)))
		else:
			print('Error: the number of seed particle component names (seed_name in model variables input file) and the number of seed particle component dissociation constants (seed_diss in model variables input file) are inconsistent, please see README for guidance')
			sys.exit()

	if (len(partit_cutoff) > 1):
		print('Length of the model variables input partit_cutoff should have a maximum length of one, but is greater, please see README for guidance')
		sys.exit()

	return(wall_on, pconc, lowsize, std, mean_rad, new_partr, chamR, chem_sch_mark, af_path, int_tol, update_stp, tot_time, siz_str, light_stat, light_time, seedVr, seed_diss, uman_up)
